import numpy as np

from nautilus import neural


def test_neural_network_emulator():
    # Test that the emulator works and produces decent estimates.

    np.random.seed(0)
    n_dim, n_points = 5, 1000
    x = np.random.random((n_points, n_dim))
    y = np.linalg.norm(x - 0.5, axis=1)
    y = np.argsort(np.argsort(y)) / float(len(y))
    emu = neural.NeuralNetworkEmulator.train(x, y, n_networks=1, pool=None)
    assert np.sqrt(np.mean((y - emu.predict(x))**2)) < 0.3 * np.std(y)



"""Module implementing unions of basic bounds."""

import itertools
import numpy as np
from scipy.special import logsumexp
from scipy.optimize import minimize
from sklearn.mixture import GaussianMixture

from .basic import UnitCube, Ellipsoid, UnitCubeEllipsoidMixture


def ellipsoids_overlap(ellipsoids):
    """Determine if ellipsoids overlap.

    This functions is based on ieeexplore.ieee.org/document/6289830.

    Parameters
    ----------
    ellipsoids : list
        List of ellipsoids.

    Returns
    -------
    overlapping : bool
        True, if there is overlap between ellipoids and False otherwise.

    """
    c = [ellipsoid.c for ellipsoid in ellipsoids]
    A_inv = [np.linalg.inv(ellipsoid.A) for ellipsoid in ellipsoids]

    for i_1, i_2 in itertools.combinations(range(len(c)), 2):
        d = c[i_1] - c[i_2]
        def k(s): return (1 - np.dot(np.dot(
            d, np.linalg.inv(A_inv[i_1] / (1 - s) + A_inv[i_2] / s)), d))
        if minimize(k, 0.5, bounds=[(1e-9, 1-1e-9)]).fun > 0:
            return True

    return False


class Union():
    r"""Union of multiple ellipsoids or unit cube-ellipsoid mixtures.

    Attributes
    ----------
    n_dim : int
        Number of dimensions.
    enlarge_per_dim : float
        Along each dimension, ellipsoids enlarged by this factor.
    n_points_min : int or None
        The minimum number of points each ellipsoid should have. Effectively,
        ellipsoids with less than twice that number will not be split further.
    cube : UnitCube or None
        If not None, the bound is defined as the overlap with the unit cube.
    points_bounds : list
        The points used to create the individual bounds. Used to split bounds
        further.
    bounds : list
        List of individual bounds.
    log_v : numpy.ndarray
        Natural log of the volume of each individual bound.
    block : numpy.ndarray
        List indicating whether a bound should not be split further because
        the volume would not decrease or the number of points is too low.
    points : numpy.ndarray
        Points that a call to `sample` will return next.
    n_sample : int
        Number of points sampled from all bounds.
    n_reject : int
        Number of points rejected due to overlap.
    rng : numpy.random.Generator
        Determines random number generation.

    """

    @classmethod
    def compute(cls, points, enlarge_per_dim=1.1, n_points_min=None,
                unit=True, bound_class=Ellipsoid, rng=None):
        """Compute the bound.

        Upon creation, the bound consists of a single individual bound.

        Parameters
        ----------
        points : numpy.ndarray with shape (n_points, n_dim)
            A 2-D array where each row represents a point.
        enlarge_per_dim : float, optional
            Along each dimension, the ellipsoid is enlarged by this factor.
            Default is 1.1.
        n_points_min : int or None, optional
            The minimum number of points each ellipsoid should have.
            Effectively, ellipsoids with less than twice that number will not
            be split further. If None, uses `n_points_min = n_dim + 1`. Default
            is None.
        unit : bool, optional
            If the bound is restricted to the overlap with the unit cube.
        bound_class : class, optional
            Type of the individual bounds, i.e. ellipsoids or unit
            cube-ellipsoid mixtures.
        rng : None or numpy.random.Generator, optional
            Determines random number generation. Default is None.

        Raises
        ------
        ValueError
            If `n_points_min` is smaller than the number of dimensions plus
            one.

        Returns
        -------
        bound : Union
            The bound.

        """
        bound = cls()
        bound.n_dim = points.shape[1]
        bound.enlarge_per_dim = enlarge_per_dim

        if n_points_min is None:
            bound.n_points_min = bound.n_dim + 1
        else:
            if n_points_min < bound.n_dim + 1:
                raise ValueError('The number of points per bound must be ' +
                                 'larger than the number of dimensions.')
            bound.n_points_min = n_points_min

        if not unit:
            bound.cube = None
        else:
            bound.cube = UnitCube.compute(
                bound.n_dim, rng=rng)

        bound.points_bounds = [points]
        bound.bounds = [bound_class.compute(
            points, enlarge_per_dim=enlarge_per_dim,
            rng=rng)]
        bound.log_v = np.array([bound.bounds[0].volume()])
        bound.block = np.atleast_1d(len(points) < 2 * bound.n_points_min)

        bound.points = np.zeros((0, points.shape[1]))
        bound.n_sample = 0
        bound.n_reject = 0

        if rng is None:
            bound.rng = np.random.default_rng()
        else:
            bound.rng = rng

        return bound

    def split_bound(self, allow_overlap=True):
        """Split the largest bound in the union.

        Parameters
        ----------
        allow_overlap : bool, optional
            Whether to allow splitting the largest bound if doing so creates
            overlaps between bounds. Cannot be False if the individual bounds
            are cube-ellipsoid mixtures. Default is True.

        Raises
        ------
        ValueError
            If `allow_overlap` is False and the individual bounds are
            cube-ellipsoid mixtures.

        Returns
        -------
        success : bool
            Whether it was possible to split any ellipsoid.

        """
        if not allow_overlap and not isinstance(self.bounds[0], Ellipsoid):
            raise ValueError("'allow_overlap' can only be False if " +
                             "bounds are ellipsoids.")

        if not np.any(~self.block):
            return False

        index = np.argmax(np.where(~self.block, self.log_v, -np.inf))
        points = self.bounds[index].transform(self.points_bounds[index])

        labels = GaussianMixture(
            n_components=2, n_init=10, random_state=self.rng.integers(
                2**32 - 1)).fit_predict(points)

        if not np.all(np.bincount(labels) >= self.n_points_min):
            self.block[index] = True
            return self.split_bound(allow_overlap=allow_overlap)

        new_bounds = []
        points = self.points_bounds[index]
        for label in [0, 1]:
            new_bounds.append(type(self.bounds[0]).compute(
                points[labels == label], enlarge_per_dim=self.enlarge_per_dim,
                rng=self.rng))

        if not allow_overlap and ellipsoids_overlap(
                self.bounds[:index] + self.bounds[index+1:] + new_bounds):
            return False

        if (logsumexp([new_bounds[0].volume(), new_bounds[1].volume()]) >
                self.bounds[index].volume()):
            self.block[index] = True
            return self.split_bound(allow_overlap=allow_overlap)

        self.points_bounds.pop(index)
        self.points_bounds.append(points[labels == 0])
        self.points_bounds.append(points[labels == 1])
        self.bounds.pop(index)
        self.bounds = self.bounds + new_bounds
        self.log_v = np.array([bound.volume() for bound in self.bounds])
        self.block = np.concatenate(
            (np.delete(self.block, index),
             [len(self.points_bounds[-2]) < 2 * self.n_points_min,
              len(self.points_bounds[-1]) < 2 * self.n_points_min]))

        # Reset the sampling.
        self.reset()

        return True

    def contains(self, points):
        """Check whether points are contained in the bound.

        Parameters
        ----------
        points : numpy.ndarray
            A 1-D or 2-D array containing single point or a collection of
            points. If more than one-dimensional, each row represents a point.

        Returns
        -------
        in_bound : bool or numpy.ndarray
            Bool or array of bools describing for each point whether it is
            contained in the bound.

        """
        in_bound = np.any([bound.contains(points) for bound in self.bounds],
                          axis=0)
        if self.cube is not None:
            in_bound = in_bound & self.cube.contains(points)
        return in_bound

    def sample(self, n_points=100):
        """Sample points from the bound.

        Parameters
        ----------
        n_points : int, optional
            How many points to draw. Default is 100.

        Returns
        -------
        points : numpy.ndarray
            Points as two-dimensional array of shape (n_points, n_dim).

        """
        while len(self.points) < n_points:
            n_sample = 1000

            p = np.exp(np.array(self.log_v) - logsumexp(self.log_v))
            n_per_bound = self.rng.multinomial(n_sample, p)

            points = np.vstack([bound.sample(n) for bound, n in
                                zip(self.bounds, n_per_bound)])
            if self.cube is not None:
                points = points[self.cube.contains(points)]
            self.rng.shuffle(points)
            n_bound = np.sum([bound.contains(points) for bound in self.bounds],
                             axis=0)
            p = 1 - 1.0 / n_bound
            points = points[self.rng.random(size=len(points)) > p]
            self.points = np.vstack([self.points, points])

            self.n_sample += n_sample
            self.n_reject += n_sample - len(points)

        points = self.points[:n_points]
        self.points = self.points[n_points:]
        return points

    def volume(self):
        """Return the natural log of the volume of the bound.

        Returns
        -------
        log_v : float
            The natural log of the volume.

        """
        if self.n_sample == 0:
            self.sample()

        return logsumexp(self.log_v) + np.log(
            1.0 - self.n_reject / self.n_sample)

    def write(self, group):
        """Write the bound to an HDF5 group.

        Parameters
        ----------
        group : h5py.Group
            HDF5 group to write to.

        """
        group.attrs['type'] = 'MultiEllipsoid'
        for key in ['n_dim', 'log_v', 'enlarge_per_dim', 'n_points_min',
                    'n_sample', 'n_reject']:
            group.attrs[key] = getattr(self, key)

        group.attrs['unit'] = self.cube is not None
        if self.cube is not None:
            self.cube.write(group.create_group('cube'))

        group.attrs['bound_class'] = self.bounds[0].__class__.__name__
        for i, bound in enumerate(self.bounds):
            bound.write(group.create_group('bound_{}'.format(i)))

        for i, points in enumerate(self.points_bounds):
            group.create_dataset('points_bound_{}'.format(i), data=points)
        group.create_dataset('points', data=self.points,
                             maxshape=(None, self.n_dim))

    def update(self, group):
        """Update bound information previously written to an HDF5 group.

        Parameters
        ----------
        group : h5py.Group
            HDF5 group to write to.

        """
        group.attrs['n_sample'] = self.n_sample
        group.attrs['n_reject'] = self.n_reject
        group['points'].resize(self.points.shape)
        group['points'][...] = self.points

    @classmethod
    def read(cls, group, rng=None):
        """Read the bound from an HDF5 group.

        Parameters
        ----------
        group : h5py.Group
            HDF5 group to write to.
        rng : None or numpy.random.Generator, optional
            Determines random number generation. Default is None.

        Returns
        -------
        bound : MultiEllipsoid
            The bound.

        """
        bound = cls()

        if rng is None:
            bound.rng = np.random.default_rng()
        else:
            bound.rng = rng

        for key in ['n_dim', 'log_v', 'enlarge_per_dim', 'n_points_min',
                    'n_sample', 'n_reject']:
            setattr(bound, key, group.attrs[key])

        if group.attrs['unit']:
            bound.cube = UnitCube.read(
                group['cube'], rng=bound.rng)

        if group.attrs['bound_class'] == 'Ellipsoid':
            bound_class = Ellipsoid
        else:
            bound_class = UnitCubeEllipsoidMixture

        bound.bounds = [bound_class.read(
            group['bound_{}'.format(i)], rng=bound.rng)
            for i in range(len(bound.log_v))]
        bound.points_bounds = [np.array(group['points_bound_{}'.format(i)]) for
                               i in range(len(bound.log_v))]
        bound.points = np.array(group['points'])

        return bound

    def reset(self, rng=None):
        """Reset random number generation and any progress, if applicable.

        Parameters
        ----------
        rng : None or numpy.random.Generator, optional
            Determines random number generation. If None, random number
            generation is not reset. Default is None.

        """
        self.points = np.zeros((0, self.n_dim))
        self.n_sample = 0
        self.n_reject = 0

        if rng is not None:
            self.rng = rng
            if self.cube is not None:
                self.cube.reset(rng)
            for bound in self.bounds:
                bound.reset(rng)



"""Module implementing neural network emulators."""

import numpy as np
import warnings
from functools import partial
from sklearn.neural_network import MLPRegressor
from threadpoolctl import threadpool_limits


@threadpool_limits.wrap(limits=1)
def train_network(x, y, neural_network_kwargs, random_state):
    """Train a network.

    Parameters
    ----------
    x : numpy.ndarray
        Input coordinates.
    y : numpy.ndarray
        Target values.
    neural_network_kwargs : dict
        Keyword arguments passed to the constructor of MLPRegressor.
    random_state : int
        Determines random number generation.

    Returns
    -------
    network : MLPRegressor
        The trained network.

    """
    return MLPRegressor(
        random_state=random_state, **neural_network_kwargs).fit(x, y)


class NeuralNetworkEmulator():
    """Likelihood neural network emulator.

    Attributes
    ----------
    mean : numpy.ndarray
        Mean of input coordinates used for normalizing coordinates.
    mean : numpy.ndarray
        Standard deviation of input coordinates used for normalizing
        coordinates.
    network : sklearn.neural_network.MLPRegressor
        Artifical neural network used for emulation.

    """

    @classmethod
    def train(cls, x, y, n_networks=4, neural_network_kwargs={}, pool=None):
        """Initialize and train the likelihood neural network emulator.

        Parameters
        ----------
        x : numpy.ndarray
            Input coordinates.
        y : numpy.ndarray
            Target values.
        n_networks : int, optional
            Number of networks used in the emulator. Default is 4.
        neural_network_kwargs : dict, optional
            Non-default keyword arguments passed to the constructor of
            MLPRegressor.
        pool : multiprocessing.Pool, optional
            Pool used for parallel processing.

        Returns
        -------
        emulator : NeuralNetworkEmulator
            The likelihood neural network emulator.

        """
        emulator = cls()

        emulator.mean = np.mean(x, axis=0)
        emulator.scale = np.std(x, axis=0)

        default_neural_network_kwargs = dict(
            hidden_layer_sizes=(100, 50, 20), alpha=0, learning_rate_init=1e-2,
            max_iter=10000, tol=0, n_iter_no_change=10)
        default_neural_network_kwargs.update(neural_network_kwargs)
        neural_network_kwargs = default_neural_network_kwargs

        if 'random_state' in neural_network_kwargs:
            warnings.warn("The 'random_state' keyword argument passed to the" +
                          " neural network is ignored.", Warning, stacklevel=2)
            del neural_network_kwargs['random_state']

        f = partial(train_network, (x - emulator.mean) / emulator.scale, y,
                    neural_network_kwargs)

        if pool is None:
            emulator.neural_networks = list(map(f, range(n_networks)))
        else:
            emulator.neural_networks = pool.map(f, range(n_networks))

        return emulator

    def predict(self, x):
        """Calculate the emulator likelihood prediction for a group of points.

        Parameters
        ----------
        x : numpy.ndarray
            Normalized coordinates of the training points.

        Returns
        -------
        y_emu : numpy.ndarray
            Emulated normalized likelihood value of the training points.

        """
        return np.mean(
            [network.predict((x - self.mean) / self.scale) for network in
             self.neural_networks], axis=0)

    def write(self, group):
        """Write the emulator to an HDF5 group.

        Parameters
        ----------
        group : h5py.Group
            HDF5 group to write to.

        """
        group.attrs['n_networks'] = len(self.neural_networks)

        for i, network in enumerate(self.neural_networks):

            for key in network.__dict__:
                if key in ['coefs_', 'intercepts_']:
                    continue
                try:
                    group.attrs[key + '_{}'.format(i)] = getattr(network, key)
                except (TypeError, ValueError):
                    pass

            for k in range(network.n_layers_ - 1):
                group.create_dataset('coefs_{}_{}'.format(k, i),
                                     data=network.coefs_[k])
                group.create_dataset('intercepts_{}_{}'.format(k, i),
                                     data=network.intercepts_[k])

        group.create_dataset('mean', data=self.mean)
        group.create_dataset('scale', data=self.scale)

    @classmethod
    def read(cls, group):
        """Read the emulator from an HDF5 group.

        Parameters
        ----------
        group : h5py.Group
            HDF5 group to write to.

        Returns
        -------
        emulator : NeuralNetworkEmulator
            The likelihood neural network emulator.

        """
        emulator = cls()

        emulator.mean = np.array(group['mean'])
        emulator.scale = np.array(group['scale'])

        emulator.neural_networks = []

        for i in range(group.attrs['n_networks']):

            network = MLPRegressor()

            for key in group.attrs:
                if key.rsplit('_', 1)[1] == '{}'.format(i):
                    setattr(network, key.rsplit('_', 1)[0], group.attrs[key])

            network.coefs_ = [
                np.array(group['coefs_{}_{}'.format(k, i)]) for k in
                range(network.n_layers_ - 1)]
            network.intercepts_ = [
                np.array(group['intercepts_{}_{}'.format(k, i)]) for k in
                range(network.n_layers_ - 1)]

            emulator.neural_networks.append(network)

        return emulator



import numpy as np
import pytest
import warnings

from scipy.stats import multivariate_normal, norm

from nautilus import Prior, Sampler


@pytest.mark.parametrize("n_networks", [0, 1, 2])
@pytest.mark.parametrize("vectorized", [True, False])
@pytest.mark.parametrize("pass_dict", [True, False])
def test_sampler_basic(n_networks, vectorized, pass_dict):
    # Test that the sampler does not crash.

    if pass_dict:
        prior = Prior()
        prior.add_parameter('a')
        prior.add_parameter('b')
    else:
        def prior(x):
            return x

    def likelihood(x):
        if pass_dict:
            x = np.squeeze(np.column_stack([x['a'], x['b']]))
        return -np.linalg.norm(x - 0.5, axis=-1) * 0.001

    sampler = Sampler(
        prior, likelihood, n_dim=2, n_networks=n_networks,
        vectorized=vectorized, pass_dict=pass_dict, n_live=500)
    sampler.run(f_live=0.45, n_eff=0, verbose=False)
    points, log_w, log_l = sampler.posterior(return_as_dict=pass_dict)
    assert sampler.effective_sample_size() > 0
    sampler.evidence()
    assert sampler.asymptotic_sampling_efficiency() > 0
    assert sampler.asymptotic_sampling_efficiency() < 1


@pytest.mark.parametrize("discard_exploration_start", [True, False])
@pytest.mark.parametrize("discard_exploration_end", [True, False, 1])
def test_sampler_switch_exploration(
        discard_exploration_start, discard_exploration_end):
    # Test that we can switch later whether to discard the exploration phase.

    def prior(x):
        return x

    def likelihood(x):
        return -np.linalg.norm(x - 0.5, axis=-1) * 0.001

    sampler = Sampler(
        prior, likelihood, n_dim=2, n_networks=1, vectorized=True, n_live=500)
    sampler.run(f_live=0.45, n_eff=10000, verbose=False,
                discard_exploration=discard_exploration_start)
    assert sampler.discard_exploration == discard_exploration_start
    points, log_w, log_l = sampler.posterior()
    n_start = len(points)
    log_z_start = sampler.evidence()

    if not isinstance(discard_exploration_end, bool):
        with pytest.raises(ValueError):
            sampler.discard_exploration = discard_exploration_end
        return

    sampler.discard_exploration = discard_exploration_end
    points, log_w, log_l = sampler.posterior()
    n_end = len(points)
    log_z_end = sampler.evidence()

    assert ((discard_exploration_start == discard_exploration_end) ==
            (n_start == n_end))
    assert ((discard_exploration_start == discard_exploration_end) ==
            (log_z_start == log_z_end))


@pytest.mark.parametrize("custom_prior", [True, False])
@pytest.mark.parametrize("vectorized", [True, False])
@pytest.mark.parametrize("pass_dict", [True, False])
def test_sampler_prior(custom_prior, vectorized, pass_dict):
    # Test that the sampler can handle all prior defintions.

    if custom_prior:
        if pass_dict:
            def prior(x):
                return dict(a=x[..., 0], b=x[..., 1])
        else:
            def prior(x):
                return x
    else:
        prior = Prior()
        prior.add_parameter('a')
        prior.add_parameter('b')

    def likelihood(x):
        if isinstance(x, dict):
            x = np.squeeze(np.column_stack([x['a'], x['b']]))
        return -np.linalg.norm(x - 0.5, axis=-1) * 0.001

    sampler = Sampler(
        prior, likelihood, n_dim=2, n_networks=1, vectorized=vectorized,
        pass_dict=pass_dict, n_live=500)
    sampler.run(f_live=0.45, n_eff=0, verbose=False)
    points, log_w, log_l = sampler.posterior(return_as_dict=pass_dict)
    if custom_prior and pass_dict:
        with pytest.raises(ValueError):
            points, log_w, log_l = sampler.posterior(return_as_dict=False)


@pytest.mark.parametrize("n_networks", [0, 4])
@pytest.mark.parametrize("discard_exploration", [True, False])
def test_sampler_accuracy(n_networks, discard_exploration):
    # Do a basic tests of the sampler accuracy.

    n_dim = 2
    mean = np.repeat(0.5, n_dim)
    cov = np.eye(n_dim) * 0.01

    def prior(x):
        return x

    def likelihood(x):
        return multivariate_normal.logpdf(x, mean=mean, cov=cov)

    sampler = Sampler(prior, likelihood, n_dim=n_dim, n_live=500,
                      n_networks=n_networks, seed=0)
    sampler.run(discard_exploration=discard_exploration, f_live=0.1,
                verbose=False)

    assert np.abs(sampler.evidence()) < 0.05

    for equal_weight in [True, False]:
        points, log_w, log_l = sampler.posterior(equal_weight=equal_weight)
        weights = np.exp(log_w)

        mean_sampler = np.average(points, weights=weights, axis=0)
        cov_sampler = np.cov(points, aweights=weights, rowvar=False)
        assert np.all(np.isclose(mean_sampler, mean, atol=0.005, rtol=0))
        assert np.all(np.isclose(cov_sampler, cov, atol=0.001, rtol=0))

    # This is a simple problem so the bounds should be perfectly nested, i.e.
    # bound i+1 is fully contained within bound i.
    shell_bound_occupation = sampler.shell_bound_occupation()
    assert np.all(shell_bound_occupation ==
                  np.tril(np.ones_like(shell_bound_occupation)))


def test_sampler_enlarge_per_dim():
    # Test that the enlarge_per_dim keyword is passed correctly. In this
    # example, we choose an enlargment factor so big that every bound should be
    # the same if we don't use neural networks. Thus, after the exploration
    # phase, all but one bound should be left.

    def prior(x):
        return x

    def likelihood(x):
        return -np.linalg.norm(x - 0.5)**2 * 0.001

    sampler = Sampler(prior, likelihood, n_dim=2, enlarge_per_dim=100,
                      n_networks=0, seed=0)
    sampler.run(f_live=0.1, n_eff=0)

    # The effective sample size should be very close to the number of calls
    # since the likelihood is extremely flat.
    assert np.isclose(sampler.n_like, sampler.effective_sample_size(), rtol=0,
                      atol=1)
    # Only one bound should be left.
    assert len(sampler.bounds) == 1
    # Check evidence accuracy.
    assert np.isclose(sampler.evidence(), -4 * 0.5**3 / 3 * 0.001, rtol=0,
                      atol=1e-4)


def test_sampler_funnel():
    # Test the sampler on a funnel distribution. This is a great, challenging
    # distribution. Also, the nature of the likelihood leads to nautilus
    # boundaries that are not strictly nested, unlike for simpler
    # distributions.

    def prior(x):
        return x

    def likelihood(x):
        return (norm.logpdf(x[0], loc=0.5, scale=0.1) +
                norm.logpdf(x[1], loc=0.5, scale=np.exp(20 * (x[0] - 0.5)) /
                            100))

    # Determine the true evidence. The likelihood is normalized so it should be
    # close to 1 minus a correction for the fraction of points falling outside
    # the unit cube.
    np.random.seed(0)
    x_0 = np.random.normal(loc=0.5, scale=0.1, size=1000000)
    x_1 = np.random.normal(loc=0.5, scale=np.exp(20 * (x_0 - 0.5)) / 100)
    log_z_true = np.log(np.mean((x_0 > 0) & (x_0 < 1) & (x_1 > 0) & (x_1 < 1)))

    sampler = Sampler(prior, likelihood, n_dim=2, n_networks=1, seed=0)
    sampler.run()
    assert np.isclose(log_z_true, sampler.evidence(), rtol=0, atol=0.1)
    # Check whether the boundaries nautilus drew are strictly nested.
    shell_bound_occupation = sampler.shell_bound_occupation()
    if np.all(shell_bound_occupation ==
              np.tril(np.ones_like(shell_bound_occupation))):
        warnings.warn('The funnel distribution was too easy.', RuntimeWarning)



"""Module implementing the prior bounds and convencience functions."""

import numbers
import numpy as np
from scipy.stats import uniform


class Prior():
    """Helper class to construct prior bounds.

    Attributes
    ----------
    keys : list
        List of model parameters.
    dists : list
        List of distributions each model parameter follows.

    """

    def __init__(self):
        """Initialize a prior without any parameters."""
        self.keys = []
        self.dists = []

    def add_parameter(self, key=None, dist=(0, 1)):
        """Add a model parameter to the prior.

        Parameters
        ----------
        key : str or None
            Name of the model parameter. If None, the key name will be `x_i`,
            where i is a number.
        dist : float, tuple, str or object
            Distribution the parameter should follow. If a float, the parameter
            is fixed to this value and will not be fitted in any analysis. If
            a tuple, it gives the lower and upper bound of a uniform
            distribution. If a string, the parameter will always be equal
            to the named model parameter. Finally, if an object, it must have
            a `isf` attribute, i.e. the inverse survival function.

        Raises
        ------
        TypeError
            If `key` or `dist` is not the correct type.
        ValueError
            If a new key already exists in the key list or if `dist` is a
            string but does not refer to a previously defined key.
        """
        if key is None:
            self.keys.append('x_{}'.format(len(self.keys)))
        elif not isinstance(key, str):
            raise TypeError("Keyword argument 'key' must be a string.")
        elif key in self.keys:
            raise ValueError("Key '{}' already in key list.".format(key))
        else:
            self.keys.append(key)

        if isinstance(dist, tuple):
            self.dists.append(uniform(loc=dist[0], scale=dist[1] - dist[0]))
        elif isinstance(dist, numbers.Number) or hasattr(dist, 'isf'):
            self.dists.append(dist)
        elif isinstance(dist, str):
            if dist not in self.keys or dist == str(key):
                raise ValueError('Key {} not defined previously.'.format(dist))
            while isinstance(self.dists[self.keys.index(dist)], str):
                dist = self.dists[self.keys.index(dist)]
            self.dists.append(dist)
        else:
            raise TypeError("Keyword argument 'dist' does not have the " +
                            "correct type")

    def dimensionality(self):
        """Determine the number of free model parameters.

        Returns
        -------
        n_dim : int
            The number of free model parameters.
        """
        return sum(not isinstance(dist, (numbers.Number, str)) for dist in
                   self.dists)

    def unit_to_physical(self, points):
        """Convert points from the unit hypercube to physical parameters.

        Parameters
        ----------
        points : numpy.ndarray
            A 1-D or 2-D array containing single point or a collection of
            points in the unit hypercube. If more than one-dimensional, each
            row represents a point.

        Returns
        -------
        phys_points : numpy.ndarray
            Points transferred into the prior volume. Has the same shape as
            `points`.

        Raises
        ------
        ValueError
            If dimensionality of `points` does not match the prior.

        """
        phys_points = np.zeros_like(points)

        try:
            assert self.dimensionality() == points.shape[-1]
        except AssertionError:
            raise ValueError('Dimensionality of points does not match prior.')

        i = 0
        for dist in self.dists:
            if hasattr(dist, 'isf'):
                phys_points[..., i] = dist.isf(1 - points[..., i])
                i = i + 1

        return phys_points

    def physical_to_dictionary(self, phys_points):
        """Convert points in the physical parameter space as a dictionary.

        Parameters
        ----------
        phys_points : numpy.ndarray
            Points in the physical parameter space. If more than
            one-dimensional, each row represents a point.

        Returns
        -------
        param_dict : dict
            Points as a dictionary. Each model parameter, including fixed ones,
            can be accessed via their key.

        Raises
        ------
        ValueError
            If dimensionality of `phys_points` does not match the prior.

        """
        try:
            assert self.dimensionality() == phys_points.shape[-1]
        except AssertionError:
            raise ValueError('Dimensionality of points does not match prior.')

        param_dict = {}

        i = 0
        for key, dist in zip(self.keys, self.dists):
            if hasattr(dist, 'isf'):
                param_dict[key] = phys_points[..., i]
                i = i + 1
            elif isinstance(dist, numbers.Number):
                param_dict[key] = np.ones(phys_points[..., 0].shape) * dist

        for key, dist in zip(self.keys, self.dists):
            if isinstance(dist, str):
                param_dict[key] = param_dict[dist]

        return param_dict

    def unit_to_dictionary(self, points):
        """Convert points from the unit hypercube to a dictionary.

        Parameters
        ----------
        points : numpy.ndarray
            A 1-D or 2-D array containing single point or a collection of
            points in the unit hypercube. If more than one-dimensional, each
            row represents a point.

        Returns
        -------
        param_dict : dict
            Points as a dictionary. Each model parameter, including fixed ones,
            can be accessed via their key.

        """
        return self.physical_to_dictionary(self.unit_to_physical(points))



"""Module implementing the Nautilus sampler."""

try:
    import h5py
except ImportError:
    pass
import numpy as np
import warnings

from functools import partial
from multiprocessing import Pool
from pathlib import Path
from scipy.special import logsumexp
from threadpoolctl import threadpool_limits
from tqdm import tqdm

from .bounds import UnitCube, NautilusBound


class Sampler():
    """
    A dynamic sampler built upon the framework of importance nested sampling.

    Attributes
    ----------
    prior : function or nautilus.Prior
        Prior describing the mapping of the unit hypercube to physical
        parameters.
    likelihood : function
        Function returning the natural logarithm of the likelihood.
    n_dim : int
        Number of dimensions of the likelihood function.
    n_live : int
        Number of live points.
    n_update : int
        The maximum number of additions to the live set before a new bound is
        created.
    n_like_new_bound : int
        The maximum number of likelihood calls before a new bounds is created.
    enlarge_per_dim : float
        Along each dimension, outer ellipsoidal bounds are enlarged by this
        factor.
    n_points_min : int or None
        The minimum number of points each ellipsoid should have. Effectively,
        ellipsoids with less than twice that number will not be split further.
    split_threshold: float
        Threshold used for splitting the multi-ellipsoidal bound used for
        sampling. If the volume of the bound prior enlarging is larger than
        `split_threshold` times the target volume, the multi-ellipsiodal
        bound is split further, if possible.
    n_networks : int
        Number of networks used in the estimator.
    neural_network_kwargs : dict
        Keyword arguments passed to the constructor of
        `sklearn.neural_network.MLPRegressor`.
    n_batch : int
        Number of likelihood evaluations that are performed at each step.
    vectorized : bool
        If True, the likelihood function can receive multiple input sets at
        once.
    pass_dict : bool
        If True, the likelihood function expects model parameters as
        dictionaries.
    pool_l : object
        Pool used to parallelize likelihood calls.
    pool_s : object
        Pool used to parallelize sampler calculations.
    rng : np.random.Generator
        Random number generator of the sampler.
    n_like : int
        Total number of likelihood evaluations.
    explored : bool
        Whether the space has been explored and the shells have been
        constructed.
    bounds : list
        List of all the constructed bounds.
    points : list
        List of arrays where each array at position i lists the points
        belonging to the i-th bound/shell.
    log_l : list
        Log likelihood values of each point. Same ordering as `points`.
    blobs : list
        Blobs associated with each point. Same ordering as `points`.
    blobs_dtype : numpy.dtype
        Data type of the blobs.
    _discard_exploration : bool
        Whether to exclude points in the exploration phase.
    shell_n : numpy.ndarray
        Number of points for each bound/shell.
    shell_n_sample : numpy.ndarray
        Number of points sampled in each bound.
    shell_n_eff : numpy.ndarray
        Effective sample size for each bound/shell.
    shell_log_l_min : numpy.ndarray
        Minimum logarithm of the likelihood required for an update in each
        bound/shell.
    shell_log_l : numpy.ndarray
        Logarithm of the mean likelihood of points in each bound/shell.
    shell_log_v : numpy.ndarray
        Logarithm of the volume of each bound/shell.
    shell_n_sample_exp : numpy.ndarray
        Number of points sampled in each bound during the exploration phase.
    shell_end_exp : numpy.ndarray
        Index at which points are coming from the sampling phase instead of the
        exploration phase.

    """

    def __init__(self, prior, likelihood, n_dim=None, n_live=2000,
                 n_update=None, enlarge_per_dim=1.1, n_points_min=None,
                 split_threshold=100, n_networks=4,
                 neural_network_kwargs=dict(), prior_args=[],
                 prior_kwargs=dict(), likelihood_args=[],
                 likelihood_kwargs=dict(), n_batch=100,
                 n_like_new_bound=None, vectorized=False, pass_dict=None,
                 pool=None, n_jobs=None, seed=None, blobs_dtype=None,
                 filepath=None, resume=True):
        r"""
        Initialize the sampler.

        Parameters
        ----------
        prior : function or nautilus.Prior
            Prior describing the mapping of the unit hypercube to the
            parameters.
        likelihood : function
            Function returning the natural logarithm of the likelihood.
        n_dim : int, optional
            Number of dimensions of the likelihood function. If not specified,
            it will be inferred from the `prior` argument. But this requires
            `prior` to be an instance of `nautilus.Prior`.
        n_live : int, optional
            Number of so-called live points. New bounds are constructed so that
            they encompass the live points. Default is 3000.
        n_update : None or int, optional
            The maximum number of additions to the live set before a new bound
            is created. If None, use `n_live`. Default is None.
        enlarge_per_dim : float, optional
            Along each dimension, outer ellipsoidal bounds are enlarged by this
            factor. Default is 1.1.
        n_points_min : int or None, optional
            The minimum number of points each ellipsoid should have.
            Effectively, ellipsoids with less than twice that number will not
            be split further. If None, uses `n_points_min = n_dim + 50`.
            Default is None.
        split_threshold: float, optional
            Threshold used for splitting the multi-ellipsoidal bound used for
            sampling. If the volume of the bound prior enlarging is larger than
            `split_threshold` times the target volume, the multi-ellipsiodal
            bound is split further, if possible. Default is 100.
        n_networks : int, optional
            Number of networks used in the estimator. Default is 4.
        neural_network_kwargs : dict, optional
            Non-default keyword arguments passed to the constructor of
            MLPRegressor.
        prior_args : list, optional
            List of extra positional arguments for `prior`. Only used if
            `prior` is a function.
        prior_kwargs : dict, optional
            Dictionary of extra keyword arguments for `prior`. Only used if
            `prior` is a function.
        likelihood_args : list, optional
            List of extra positional arguments for `likelihood`.
        likelihood_kwargs : dict, optional
            Dictionary of extra keyword arguments for `likelihood`.
        n_batch : int, optional
            Number of likelihood evaluations that are performed at each step.
            If likelihood evaluations are parallelized, should be multiple
            of the number of parallel processes. Very large numbers can
            lead to new bounds being created long after `n_update` additions to
            the live set have been achieved. This will not cause any bias but
            could reduce efficiency. Default is 100.
        n_like_new_bound : None or int, optional
            The maximum number of likelihood calls before a new bounds is
            created. If None, use 10 times `n_live`. Default is None.
        vectorized : bool, optional
            If True, the likelihood function can receive multiple input sets
            at once. For example, if the likelihood function receives arrays,
            it should be able to take an array with shape (n_points, n_dim)
            and return an array with shape (n_points). Similarly, if the
            likelihood function accepts dictionaries, it should be able to
            process dictionaries where each value is an array with shape
            (n_points). Default is False.
        pass_dict : bool or None, optional
            If True, the likelihood function expects model parameters as
            dictionaries. If False, it expects regular numpy arrays. Default is
            to set it to True if prior was a nautilus.Prior instance and False
            otherwise.
        pool : None, object, int or tuple, optional
            Pool used for parallelization of likelihood calls and sampler
            calculations. If None, no parallelization is performed. If an
            integer, the sampler will use a multiprocessing.Pool object with
            the specified number of processes. Finally, if specifying a tuple,
            the first one specifies the pool used for likelihood calls and the
            second one the pool for sampler calculations. Default is None.
        n_jobs : int or string, optional
            Deprecated.
        seed : None or int, optional
            Seed for random number generation used for reproducible results
            accross different runs. If None, results are not reproducible.
            Default is None.
        blobs_dtype : object or None, optional
            Object that can be converted to a data type object describing the
            blobs. If None, this will be inferred from the first blob. Default
            is None.
        filepath : string, pathlib.Path or None, optional
            Path to the file where results are saved. Must have a '.h5' or
            '.hdf5' extension. If None, no results are written. Default is
            None.
        resume : bool, optional
            If True, resume from previous run if `filepath` exists. If False,
            start from scratch and overwrite any previous file. Default is
            True.

        Raises
        ------
        ValueError
            If `prior` is a function and `n_dim` is not given or `pass_struct`
            is set to True. If the dimensionality of the problem is less than
            2.

        """
        if callable(prior):
            self.prior = partial(prior, *prior_args, **prior_kwargs)
        else:
            self.prior = prior
        self.likelihood = partial(
            likelihood, *likelihood_args, **likelihood_kwargs)

        if callable(prior):
            if n_dim is None:
                raise ValueError("When passing a function as the 'prior' " +
                                 "argument, 'n_dim' cannot be None.")
            self.n_dim = n_dim
            if pass_dict is None:
                pass_dict = False
        else:
            self.n_dim = prior.dimensionality()
            if pass_dict is None:
                pass_dict = True

        if self.n_dim <= 1:
            raise ValueError(
                'Cannot run Nautilus with less than 2 parameters.')

        self.n_live = n_live

        if n_update is None:
            n_update = n_live
        self.n_update = n_update

        if n_like_new_bound is None:
            n_like_new_bound = 10 * n_live
        self.n_like_new_bound = n_like_new_bound

        self.enlarge_per_dim = enlarge_per_dim

        if n_points_min is None:
            n_points_min = self.n_dim + 50
        self.n_points_min = n_points_min

        self.split_threshold = split_threshold

        self.n_networks = n_networks

        self.neural_network_kwargs = neural_network_kwargs
        self.n_batch = n_batch
        self.vectorized = vectorized
        self.pass_dict = pass_dict

        if pool is None or pool == 1:
            self.pool_l = None
            self.pool_s = None
        elif isinstance(pool, int):
            self.pool_l = Pool(pool)
            self.pool_s = self.pool_l
        elif isinstance(pool, tuple):
            self.pool_l = pool[0]
            if isinstance(self.pool_l, int):
                self.pool_l = Pool(self.pool_l)
            self.pool_s = pool[1]
            if isinstance(self.pool_s, int):
                self.pool_s = Pool(self.pool_s)
        else:
            self.pool_l = pool
            self.pool_s = pool

        if n_jobs is not None:
            warnings.warn(
                "The 'n_jobs' keyword argument has been deprecated .Use " +
                "'pool', instead.", DeprecationWarning, stacklevel=2)

        self.rng = np.random.default_rng(seed)

        # The following variables carry the information about the run.
        self.n_like = 0
        self.explored = False
        self.bounds = []
        self.points = []
        self.log_l = []
        self.blobs = []
        self.blobs_dtype = blobs_dtype
        self._discard_exploration = False
        self.shell_n = np.zeros(0, dtype=int)
        self.shell_n_sample = np.zeros(0, dtype=int)
        self.shell_n_eff = np.zeros(0, dtype=float)
        self.shell_log_l_min = np.zeros(0, dtype=float)
        self.shell_log_l = np.zeros(0, dtype=float)
        self.shell_log_v = np.zeros(0, dtype=float)
        self.shell_n_sample_exp = np.zeros(0, dtype=int)
        self.shell_end_exp = np.zeros(0, dtype=int)

        self.filepath = filepath
        if resume and filepath is not None and Path(filepath).exists():
            with h5py.File(filepath, 'r') as fstream:

                group = fstream['sampler']

                self.rng.bit_generator.state = dict(
                    bit_generator='PCG64',
                    state=dict(
                        state=int(group.attrs['rng_state']),
                        inc=int(group.attrs['rng_inc'])),
                    has_uint32=group.attrs['rng_has_uint32'],
                    uinteger=group.attrs['rng_uinteger'])

                for key in ['n_like', 'explored', '_discard_exploration',
                            'shell_n', 'shell_n_sample', 'shell_n_eff',
                            'shell_log_l_min', 'shell_log_l', 'shell_log_v',
                            'shell_n_sample_exp', 'shell_end_exp']:
                    setattr(self, key, group.attrs[key])

                for shell in range(len(self.shell_n)):
                    self.points.append(
                        np.array(group['points_{}'.format(shell)]))
                    self.log_l.append(
                        np.array(group['log_l_{}'.format(shell)]))
                    try:
                        self.blobs.append(
                            np.array(group['blobs_{}'.format(shell)]))
                        self.blobs_dtype = self.blobs[-1].dtype
                    except KeyError:
                        pass

                self.bounds = [
                    UnitCube.read(fstream['bound_0'], rng=self.rng), ]
                for i in range(1, len(self.shell_n)):
                    self.bounds.append(NautilusBound.read(
                        fstream['bound_{}'.format(i)], rng=self.rng))

    def run(self, f_live=0.01, n_shell=None, n_eff=10000,
            discard_exploration=False, verbose=False):
        """Run the sampler until convergence.

        Parameters
        ----------
        f_live : float, optional
            Maximum fraction of the evidence contained in the live set before
            building the initial shells terminates. Default is 0.01.
        n_shell : int, optional
            Minimum number of points in each shell. The algorithm will sample
            from the shells until this is reached. Default is the batch size of
            the sampler which is 100 unless otherwise specified.
        n_eff : float, optional
            Minimum effective sample size. The algorithm will sample from the
            shells until this is reached. Default is 10000.
        discard_exploration : bool, optional
            Whether to discard points drawn in the exploration phase. This is
            required for a fully unbiased posterior and evidence estimate.
            Default is False.
        verbose : bool, optional
            If True, print additional information. Default is False.

        """
        if not self.explored:

            if verbose:
                print('#########################')
                print('### Exploration Phase ###')
                print('#########################')
                print()

            while (self.live_evidence_fraction() > f_live or
                   len(self.bounds) == 0):
                self.add_bound(verbose=verbose)
                self.fill_bound(verbose=verbose)
                if self.filepath is not None:
                    self.write(self.filepath, overwrite=True)

            # If some shells are unoccupied in the end, remove them. They will
            # contain close to 0 volume and may never yield a point when
            # trying to sample from them.
            for shell in reversed(range(len(self.bounds))):
                if self.shell_n[shell] == 0:
                    self.bounds.pop(shell)
                    self.points.pop(shell)
                    self.log_l.pop(shell)
                    if self.blobs_dtype is not None:
                        self.blobs.pop(shell)
                    for key in ['shell_n', 'shell_n_sample', 'shell_n_eff',
                                'shell_log_l_min', 'shell_log_l',
                                'shell_log_v']:
                        setattr(self, key, np.delete(
                            getattr(self, key), shell))

            self.shell_n_sample_exp = np.copy(self.shell_n_sample)
            self.shell_end_exp = np.array([len(p) for p in self.points])
            self.explored = True

            if self.filepath is not None:
                self.write(self.filepath, overwrite=True)

        self.discard_exploration = discard_exploration
        if self.filepath is not None:
            self.write_shell_information_update(self.filepath)

        if n_shell is None:
            n_shell = self.n_batch

        if (np.any(self.shell_n < n_shell) or
                self.effective_sample_size() < n_eff):

            if verbose:
                print('#########################')
                print('##### Sampling Phase ####')
                print('#########################')
                print()

            self.add_points(n_shell=n_shell, n_eff=n_eff, verbose=verbose)

    @property
    def discard_exploration(self):
        """Return whether the exploration phase is discarded.

        Returns
        -------
        discard_exploration : bool
            Whether the exploration phase is discarded.

        """
        return self._discard_exploration

    @discard_exploration.setter
    def discard_exploration(self, discard_exploration):
        """Set whether exploration phase should be discarded.

        Parameters
        ----------
        discard_exploration : bool
            Whether the exploration phase is discarded.

        Raises
        ------
        ValueError
            If `discard_exploration` is not a bool.

        """
        if not isinstance(discard_exploration, bool):
            raise ValueError("'discard_exploration' must be a bool.")

        self._discard_exploration = discard_exploration
        for index in range(len(self.log_l)):
            self.update_shell_info(index)

    def posterior(self, return_as_dict=None, equal_weight=False,
                  return_blobs=False):
        """Return the posterior sample estimate.

        Parameters
        ----------
        return_as_dict : bool or None, optional
            If True, return `points` as a dictionary. If None, will default to
            False unless one uses custom prior that only returns dictionaries.
            Default is None.
        equal_weight : bool, optional
            If True, return an equal weighted posterior. Default is False.
        return_blobs : bool, optional
            If True, return the blobs. Default is False.

        Returns
        -------
        points : numpy.ndarray or dict
            Points of the posterior.
        log_w : numpy.ndarray
            Weights of each point of the posterior.
        log_l : numpy.ndarray
            Logarithm of the likelihood at each point of the posterior.
        blobs : numpy.ndarray, optional
            Blobs for each point of the posterior. Only returned if
            `return_blobs` is True.

        Raises
        ------
        ValueError
            If `return_as_dict` or `return_blobs` are True but the sampler has
            not been run in a way that that's possible.

        """
        if return_as_dict is None:
            if callable(self.prior) and self.pass_dict:
                return_as_dict = True
            else:
                return_as_dict = False

        if self._discard_exploration and self.explored:
            start = self.shell_end_exp
        else:
            start = np.zeros(len(self.points), dtype=int)

        points = np.concatenate([p[s:] for p, s in zip(self.points, start)])
        log_v = np.repeat(self.shell_log_v -
                          np.log(np.maximum(self.shell_n, 1)), self.shell_n)
        log_l = np.concatenate([l[s:] for l, s in zip(self.log_l, start)])
        log_w = log_v + log_l
        if return_blobs:
            if self.blobs_dtype is None:
                raise ValueError('No blobs have been calculated.')
            blobs = np.concatenate([b[s:] for b, s in zip(self.blobs, start)])

        if callable(self.prior):
            transform = self.prior
        else:
            if return_as_dict:
                transform = self.prior.unit_to_dictionary
            else:
                transform = self.prior.unit_to_physical

        if not self.vectorized and callable(self.prior):
            points = np.array(list(map(transform, points)))
        else:
            points = transform(points)

        if not return_as_dict and callable(self.prior) and self.pass_dict:
            raise ValueError(
                'Cannot return points as numpy array. The prior function' +
                ' only returns dictionaries.')

        if equal_weight:
            select = (self.rng.random(len(log_w)) <
                      np.exp(log_w - np.amax(log_w)))
            points = points[select]
            log_w = np.ones(len(points)) * np.log(1.0 / np.sum(select))
            log_l = log_l[select]
            if return_blobs:
                blobs = blobs[select]

        # Normalize weights.
        log_w = log_w - logsumexp(log_w)

        if return_blobs:
            return points, log_w, log_l, blobs
        else:
            return points, log_w, log_l

    def effective_sample_size(self):
        r"""Estimate the total effective sample size :math:`N_{\rm eff}`.

        Returns
        -------
        n_eff : float
            Estimate of the total effective sample size :math:`N_{\rm eff}`.

        """
        select = self.shell_n_eff > 0
        sum_w = np.exp(self.shell_log_l + self.shell_log_v -
                       np.nanmax(self.shell_log_l + self.shell_log_v))[select]
        sum_w_sq = sum_w**2 / self.shell_n_eff[select]
        return np.sum(sum_w)**2 / np.sum(sum_w_sq)

    def evidence(self):
        r"""Estimate the global evidence :math:`\log \mathcal{Z}`.

        Returns
        -------
        log_z : float
            Estimate of the global evidence :math:`\log \mathcal{Z}`.

        """
        select = ~np.isnan(self.shell_log_l)
        return logsumexp(self.shell_log_l[select] + self.shell_log_v[select])

    def asymptotic_sampling_efficiency(self):
        r"""Estimate the asymptotic sampling efficiency :math:`\eta`.

        The asymptotic sampling efficiency is defined as
        :math:`\eta = \lim_{N_{\rm like} \to \infty} N_{\rm eff} / N_{\rm like}`.
        This is set after the exploration phase. However, the estimate will be
        updated based on what is found in the sampling phase.

        Returns
        -------
        eta : float
            Estimate of the asymptotic sampling efficiency.

        """
        shell_log_z = self.shell_log_l + self.shell_log_v
        shell_eta = self.shell_n_eff / self.shell_n
        select = ~np.isnan(self.shell_log_l)
        shell_log_z = shell_log_z[select]
        shell_eta = shell_eta[select]
        return np.exp(2 * logsumexp(shell_log_z) - 2 * logsumexp(
            shell_log_z - 0.5 * np.log(shell_eta)))

    def sample_shell(self, index, shell_t=None):
        """Sample a batch of points uniformly from a shell.

        The shell at index :math:`i` is defined as the volume enclosed by the
        bound of index :math:`i` and enclosed by not other bound of index
        :math:`k` with :math:`k > i`.

        Parameters
        ----------
        index : int
            Index of the shell.
        shell_t : np.ndarray or None, optional
            If not None, an array of shell associations of possible transfer
            points.

        Returns
        -------
        points : numpy.ndarray
            Array of shape (n_shell, n_dim) containing points sampled uniformly
            from the shell.
        n_bound : int
            Number of points drawn within the bound at index :math:`i`. Will
            be different from `n_shell` if there are bounds with index
            :math:`k` with :math:`k > i`.
        idx_t : np.ndarray, optional
            Indeces of the transfer candidates that should be transferred. Only
            returned if `shell_t` is not None.

        """
        if shell_t is not None and index not in [-1, len(self.bounds) - 1]:
            raise ValueError("'shell_t' must be empty list if not sampling " +
                             "from the last bound/shell.")

        n_bound = 0
        n_sample = 0
        idx_t = np.zeros(0, dtype=int)
        points_all = []

        with threadpool_limits(limits=1):
            while n_sample < self.n_batch:
                points = self.bounds[index].sample(
                    self.n_batch - n_sample, pool=self.pool_s)
                n_bound += self.n_batch - n_sample

                # Remove points that are actually in another shell.
                in_shell = np.ones(len(points), dtype=bool)
                for bound in self.bounds[index:][1:]:
                    in_shell = in_shell & ~bound.contains(points)
                    if np.all(~in_shell):
                        continue
                points = points[in_shell]

                # Replace points for which we can use transfer points.
                replace = np.zeros(len(points), dtype=bool)
                if shell_t is not None and len(shell_t) > 0:
                    shell_p = self.shell_association(
                        points, n_max=len(self.bounds) - 1)
                    for shell in range(len(self.bounds) - 1):
                        idx_1 = np.flatnonzero(shell_t == shell)
                        idx_2 = np.flatnonzero(shell_p == shell)
                        n = min(len(idx_1), len(idx_2))
                        if n > 0:
                            idx_t = np.append(idx_t, self.rng.choice(
                                idx_1, size=n, replace=False))
                            shell_t[idx_t] = -1
                            replace[self.rng.choice(
                                idx_2, size=n, replace=False)] = True

                points = points[~replace]

                if len(points) > 0:
                    points_all.append(points)
                    n_sample += len(points)

        points = np.concatenate(points_all)

        if shell_t is None:
            return points, n_bound
        else:
            return points, n_bound, idx_t

    def evaluate_likelihood(self, points):
        """Evaluate the likelihood for a given set of points.

        Parameters
        ----------
        points : numpy.ndarray
            Points at which to evaluate the likelihood.

        Returns
        -------
        log_l : numpy.ndarray
            Natural log of the likelihood of each point.
        blobs : list, dict or None
            Blobs associated with the points, if any.

        Raises
        ------
        ValueError
            If `self.blobs_dtype` is not None but the likelihood function does
            not return blobs.

        """
        # Transform points from the unit cube to the input arguments of the
        # likelihood function.
        if callable(self.prior):
            transform = self.prior
        else:
            if self.pass_dict:
                transform = self.prior.unit_to_dictionary
            else:
                transform = self.prior.unit_to_physical

        if not self.vectorized:
            args = list(map(transform, points))
        else:
            args = transform(points)

        # Evaluate the likelihood.
        if self.vectorized:
            result = self.likelihood(args)
            if isinstance(result, tuple):
                result = list(zip(*result))
        elif self.pool_l is not None:
            result = list(self.pool_l.map(self.likelihood, args))
        else:
            result = list(map(self.likelihood, args))

        if isinstance(result[0], tuple):
            log_l = np.array([r[0] for r in result])
            blobs = [r[1:] for r in result]
            if self.blobs_dtype is None:
                if len(blobs[0]) > 1:
                    self.blobs_dtype = [
                        ('blob_{}'.format(i), np.array([b]).dtype) for i, b in
                        enumerate(blobs[0])]
                else:
                    self.blobs_dtype = np.array([blobs[0][0]]).dtype
            blobs = np.squeeze(np.array(blobs, dtype=self.blobs_dtype))
        else:
            log_l = np.array(result)
            blobs = None

        if blobs is None and self.blobs_dtype is not None:
            raise ValueError("'blobs_dtype' was specified but the likelihood" +
                             " function does not return blobs.")

        self.n_like += len(log_l)

        return log_l, blobs

    def update_shell_info(self, index):
        """Update the shell information for calculation of summary statistics.

        Parameters
        ----------
        index: int
            Index of the shell.

        """
        shell_n_sample = self.shell_n_sample[index]

        if self._discard_exploration and self.explored:
            start = self.shell_end_exp[index]
            shell_n_sample -= self.shell_n_sample_exp[index]
        else:
            start = 0

        log_l = self.log_l[index][start:]
        self.shell_n[index] = len(log_l)

        if self.shell_n[index] > 0:
            self.shell_log_v[index] = (
                self.bounds[index].volume() +
                np.log(self.shell_n[index] / shell_n_sample))
            self.shell_log_l[index] = logsumexp(log_l) - np.log(len(log_l))
            if not np.all(log_l == -np.inf):
                self.shell_n_eff[index] = np.exp(2 * logsumexp(log_l) -
                                                 logsumexp(2 * log_l))
            else:
                self.shell_n_eff[index] = len(log_l)
        else:
            self.shell_log_v[index] = -np.inf
            self.shell_log_l[index] = np.nan
            self.shell_n_eff[index] = 0

    def print_status(self):
        """Print current summary statistics."""
        print('N_like: {:>17}'.format(self.n_like))
        print('N_eff: {:>18.0f}'.format(self.effective_sample_size()))
        print('log Z: {:>18.3f}'.format(self.evidence()))
        if not self.explored:
            print('log V: {:>18.3f}'.format(self.shell_log_v[-1]))
            print('f_live: {:>17.3f}'.format(self.live_evidence_fraction()))

    def add_bound(self, verbose=False):
        """Try building a new bound from existing points.

        If the new bound would be larger than the previious bound, we will
        reject the new bound.

        Parameters
        ----------
        verbose : bool, optional
            If True, print additional information. Default is False.

        """
        success = False
        # If this is the first bound, use the UnitCube bound.
        if len(self.bounds) == 0:
            log_l_min = -np.inf
            self.bounds.append(UnitCube.compute(self.n_dim, rng=self.rng))
            success = True
        else:
            if verbose:
                print('Adding Bound {}'.format(len(self.bounds) + 1), end='\r')
            log_l = np.concatenate(self.log_l)
            points = np.concatenate(self.points)[np.argsort(log_l)]
            log_l = np.sort(log_l)
            log_l_min = 0.5 * (log_l[-self.n_live] + log_l[-self.n_live - 1])
            with threadpool_limits(limits=1):
                bound = NautilusBound.compute(
                    points, log_l, log_l_min, self.live_volume(),
                    enlarge_per_dim=self.enlarge_per_dim,
                    n_points_min=self.n_points_min,
                    split_threshold=self.split_threshold,
                    n_networks=self.n_networks,
                    neural_network_kwargs=self.neural_network_kwargs,
                    pool=self.pool_s, rng=self.rng)
                bound.sample(1000, return_points=False, pool=self.pool_s)
            if bound.volume() < self.bounds[-1].volume():
                self.bounds.append(bound)
                success = True

        if success:
            self.shell_n = np.append(self.shell_n, 0)
            self.shell_n_sample = np.append(self.shell_n_sample, 0)
            self.shell_n_eff = np.append(self.shell_n_eff, 0)
            self.shell_log_l = np.append(self.shell_log_l, np.nan)
            self.shell_log_v = np.append(self.shell_log_v, np.nan)
            self.shell_log_l_min = np.append(self.shell_log_l_min, log_l_min)
            self.points.append(np.zeros((0, self.n_dim)))
            self.log_l.append(np.zeros(0))
            self.blobs.append(np.zeros(0))
        else:
            self.shell_log_l_min[-1] = log_l_min

        if verbose:
            print('Adding Bound {:<5} {:>6}'.format(
                str(len(self.bounds) + (not success)) + ':', 'done' if success
                else 'failed'))
            if isinstance(self.bounds[-1], NautilusBound):
                n_neural, n_ell =\
                    self.bounds[-1].number_of_networks_and_ellipsoids()
            else:
                n_neural, n_ell = 0, 0
            if success:
                print("Ellipsoids: {:>13}".format(n_ell))
                print("Neural Networks: {:>8}".format(n_neural))

    def fill_bound(self, verbose=False):
        """Fill a new bound with points until a new bound should be created.

        Parameters
        ----------
        verbose : bool, optional
            If True, print additional information. Default is False.

        """
        self.points[-1] = [self.points[-1], ]
        self.log_l[-1] = [self.log_l[-1], ]
        if len(self.blobs[-1]) > 0:
            self.blobs[-1] = [self.blobs[-1], ]
        else:
            self.blobs[-1] = []

        shell_t = []
        points_t = []
        log_l_t = []
        blobs_t = []

        # Check which points points from previous shells could be transferred
        # to the new bound.
        if len(self.bounds) > 1:
            for shell in range(len(self.bounds) - 1):

                in_bound = self.bounds[-1].contains(self.points[shell])
                shell_t.append(np.repeat(shell, np.sum(in_bound)))

                points_t.append(self.points[shell][in_bound])
                self.points[shell] = self.points[shell][~in_bound]

                log_l_t.append(self.log_l[shell][in_bound])
                self.log_l[shell] = self.log_l[shell][~in_bound]

                if self.blobs_dtype is not None:
                    blobs_t.append(self.blobs[shell][in_bound])
                    self.blobs[shell] = self.blobs[shell][~in_bound]

                self.shell_n[shell] -= np.sum(in_bound)
                self.update_shell_info(shell)

            shell_t = np.concatenate(shell_t)
            points_t = np.concatenate(points_t)
            log_l_t = np.concatenate(log_l_t)
            if self.blobs_dtype is not None:
                blobs_t = np.concatenate(blobs_t)

        log_l_min = self.shell_log_l_min[-1]
        n_update = 0
        n_like = 0
        n_update_max = self.n_update
        n_like_max = self.n_like_new_bound
        if len(self.bounds) == 1:
            n_update_max += self.n_live
            n_like_max = np.inf

        if verbose:
            pbar = tqdm(desc='Filling Bound {}'.format(len(self.bounds)),
                        total=n_update_max, leave=False)

        while n_update < n_update_max and n_like < n_like_max:
            points, n_bound, idx_t = self.sample_shell(-1, shell_t)
            assert len(points) + len(idx_t) == n_bound
            log_l, blobs = self.evaluate_likelihood(points)
            self.points[-1].append(points)
            self.log_l[-1].append(log_l)
            if self.blobs_dtype is not None:
                self.blobs[-1].append(blobs)

            if len(idx_t) > 0:
                self.points[-1].append(points_t[idx_t])
                points_t = np.delete(points_t, idx_t, axis=0)
                self.log_l[-1].append(log_l_t[idx_t])
                log_l_t = np.delete(log_l_t, idx_t)
                if self.blobs_dtype is not None:
                    self.blobs[-1].append(blobs_t[idx_t])
                    blobs_t = np.delete(blobs_t, idx_t, axis=0)
                shell_t = np.delete(shell_t, idx_t)

            self.shell_n[-1] += n_bound
            self.shell_n_sample[-1] += n_bound
            n_update += np.sum(log_l >= log_l_min)
            n_like += len(points)

            if verbose:
                pbar.update(np.sum(log_l >= log_l_min))

        self.points[-1] = np.concatenate(self.points[-1])
        self.log_l[-1] = np.concatenate(self.log_l[-1])
        if self.blobs_dtype is not None:
            self.blobs[-1] = np.concatenate(self.blobs[-1])
        else:
            self.blobs[-1] = np.zeros(0)
        self.update_shell_info(-1)

        if verbose:
            pbar.close()
            print('Filling Bound {:<6} done'.format(
                str(len(self.bounds)) + ':'))
            self.print_status()
            print('')

    def live_evidence_fraction(self):
        """Estimate the fraction of the evidence contained in the live set.

        This estimate can be used as a stopping criterion.

        Returns
        -------
        log_z : float
            Estimate of the fraction of the evidence in the live set.

        """
        if len(self.bounds) == 0:
            return 1.0
        else:
            log_v = np.repeat(
                self.shell_log_v - np.log(np.maximum(self.shell_n, 1)),
                self.shell_n)
            log_l = np.concatenate(self.log_l)
            log_w = log_v + log_l
            log_w_live = log_w[np.argsort(log_l)][-self.n_live:]
            return np.exp(logsumexp(log_w_live) - logsumexp(log_w))

    def live_volume(self):
        """Estimate the volume that is currently contained in the live set.

        Returns
        -------
        log_v : float
            Estimate of the volume in the live set.

        """
        if len(self.bounds) == 0:
            return 1.0
        else:
            log_l = np.concatenate(self.log_l)
            log_v = np.repeat(
                self.shell_log_v - np.log(np.maximum(self.shell_n, 1)),
                self.shell_n)
            log_v_live = log_v[np.argsort(log_l)][-self.n_live:]

            return logsumexp(log_v_live)

    def add_samples_to_shell(self, shell):
        """Add samples to a shell.

        The number of points added is always equal to the batch size.

        Parameters
        ----------
        shell : int
            The index of the shell for which to add points.

        """
        points, n_bound = self.sample_shell(shell)
        self.shell_n_sample[shell] += n_bound
        log_l, blobs = self.evaluate_likelihood(points)
        self.points[shell] = np.concatenate([self.points[shell], points])
        self.log_l[shell] = np.concatenate([self.log_l[shell], log_l])
        if self.blobs_dtype is not None:
            self.blobs[shell] = np.concatenate([self.blobs[shell], blobs])
        self.update_shell_info(shell)
        if self.filepath is not None:
            self.write_shell_update(self.filepath, shell)

    def add_points(self, n_eff=0, n_shell=0, verbose=False):
        """Add samples to shells.

        This function add samples to the shell until very shell has a minimum
        number of points and a minimum effective sample size is achieved.

        Parameters
        ----------
        n_eff : float, optional
            Minimum effective sample size. The algorithm will sample from the
            shells until this is reached. Default is 0.
        n_shell : int, optional
            Minimum number of points in each shell. The algorithm will sample
            from the shells until this is reached. Default is 0.
        verbose : bool, optional
            If True, print additional information. Default is False.

        """
        idx = np.flatnonzero(self.shell_n < n_shell)
        if verbose and len(idx) > 0:
            pbar = tqdm(desc='Sampling shells ', total=len(idx),
                        leave=False)

        for index in idx:
            while self.shell_n[index] < n_shell:
                self.add_samples_to_shell(index)
            if verbose:
                pbar.update(1)

        if verbose and len(idx) > 0:
            pbar.close()
            print('Sampling shells:     done')
            self.print_status()
            print('')

        d_n_eff = n_eff - self.effective_sample_size()

        if verbose and d_n_eff > 0:
            pbar = tqdm(desc='Sampling posterior', total=n_eff,
                        leave=False, initial=self.effective_sample_size(),
                        bar_format="{l_bar}{bar}{n:.0f}/{total_fmt} " +
                        "[{elapsed}<{remaining}, {rate_fmt}{postfix}]")

        while self.effective_sample_size() < n_eff:
            index = np.argmax(
                self.shell_log_l + self.shell_log_v -
                0.5 * np.log(self.shell_n) - 0.5 * np.log(self.shell_n_eff))
            n_eff_old = self.effective_sample_size()
            self.add_samples_to_shell(index)
            n_eff_new = self.effective_sample_size()
            if verbose:
                pbar.update(n_eff_new - n_eff_old)

        if verbose and d_n_eff > 0:
            pbar.close()
            print('Sampling posterior:  done')
            self.print_status()
            print('')

    def shell_association(self, points, n_max=None):
        """Determine the shells each point belongs to.

        Parameters
        ----------
        points : numpy.ndarray
            points for which to determine shell association.
        n_max : int, optional
            The maximum number of shells to consider. Effectively, this
            determines the shell association at step `n_max` in the exploration
            phase. Default is to consider all shells.

        Returns
        -------
        shell: int
            Shell association for each point.

        """
        if n_max is None:
            n_max = len(self.bounds)

        shell = np.repeat(-1, len(points))
        for i, bound in reversed(list(enumerate(self.bounds[:n_max]))):
            mask = shell >= 0
            if np.all(mask):
                break
            mask[~mask] = ~bound.contains(points[~mask])
            shell[~mask] = i

        return shell

    def shell_bound_occupation(self, fractional=True):
        """Determine how many points of each shell are also part of each bound.

        Parameters
        ----------
        fractional : bool, optional
            Whether to return the absolute or fractional dependence. Default
            is True.

        Returns
        -------
        m : numpy.ndarray
            Two-dimensional array with occupation numbers. The element at index
            :math:`(i, j)` corresponds to the occupation of points in shell
            shell :math:`i` that also belong to bound :math:`j`. If
            `fractional` is True, this is the fraction of all points in shell
            :math:`i` and otherwise it is the absolute number.

        """
        m = np.zeros((len(self.bounds), len(self.bounds)), dtype=int)

        for i, points in enumerate(self.points):
            for k, bound in enumerate(self.bounds):
                m[i, k] = np.sum(bound.contains(points))

        if fractional:
            m = m / np.diag(m)[:, np.newaxis]

        return m

    def write(self, filepath, overwrite=False):
        """Write the sampler to disk.

        Parameters
        ----------
        filepath : string or pathlib.Path
            Path to the file. Must have a '.h5' or '.hdf5' extension.
        overwrite : bool, optional
            Whether to overwrite an existing file. Default is False.

        Raises
        ------
        ValueError
            If file extension is not '.h5' or '.hdf5'.
        RuntimeError
            If file exists and `overwrite` is False.

        """
        filepath = Path(filepath)

        if filepath.suffix not in ['.h5', '.hdf5']:
            raise ValueError("File ending must '.h5' or '.hdf5'.")

        if filepath.exists():
            if not overwrite:
                raise RuntimeError(
                    "File {} already exists.".format(str(filepath)))
            else:
                filepath.unlink()

        filepath.parent.mkdir(parents=True, exist_ok=True)

        fstream = h5py.File(filepath, 'x')
        group = fstream.create_group('sampler')

        for key in ['n_dim', 'n_live', 'n_update', 'n_like_new_bound',
                    'enlarge_per_dim', 'n_points_min', 'split_threshold',
                    'n_networks', 'n_batch', 'vectorized', 'pass_dict',
                    'n_like', 'explored', '_discard_exploration', 'shell_n',
                    'shell_n_sample', 'shell_n_eff', 'shell_log_l_min',
                    'shell_log_l', 'shell_log_v', 'shell_n_sample_exp',
                    'shell_end_exp']:
            group.attrs[key] = getattr(self, key)

        for key in self.neural_network_kwargs.keys():
            group.attrs['neural_network_{}'.format(key)] =\
                self.neural_network_kwargs[key]

        for shell in range(len(self.bounds)):
            group.create_dataset(
                'points_{}'.format(shell), data=self.points[shell],
                maxshape=(None, self.n_dim))
            group.create_dataset(
                'log_l_{}'.format(shell), data=self.log_l[shell],
                maxshape=(None, ))
            if self.blobs_dtype is not None:
                maxshape = list(self.blobs[shell].shape)
                maxshape[0] = None
                group.create_dataset(
                    'blobs_{}'.format(shell), data=self.blobs[shell],
                    maxshape=tuple(maxshape))

        for i, bound in enumerate(self.bounds):
            bound.write(fstream.create_group('bound_{}'.format(i)))

        rng_state = self.rng.bit_generator.state
        group.attrs['rng_state'] = str(rng_state['state']['state'])
        group.attrs['rng_inc'] = str(rng_state['state']['inc'])
        group.attrs['rng_has_uint32'] = rng_state['has_uint32']
        group.attrs['rng_uinteger'] = rng_state['uinteger']

        fstream.close()

    def write_shell_update(self, filepath, shell):
        """Update the sampler data for a single shell.

        Parameters
        ----------
        filepath : string or pathlib.Path
            Path to the file. Must have a '.h5' or '.hdf5' extension.
        shell : int
            Shell index for which to write the upate.

        """
        fstream = h5py.File(Path(filepath), 'r+')
        group = fstream['sampler']

        for key in ['n_like', 'shell_n', 'shell_n_sample', 'shell_n_eff',
                    'shell_log_l_min', 'shell_log_l', 'shell_log_v']:
            group.attrs[key] = getattr(self, key)

        group['points_{}'.format(shell)].resize(self.points[shell].shape)
        group['points_{}'.format(shell)][...] = self.points[shell]
        group['log_l_{}'.format(shell)].resize(self.log_l[shell].shape)
        group['log_l_{}'.format(shell)][...] = self.log_l[shell]
        if self.blobs_dtype is not None:
            group['blobs_{}'.format(shell)].resize(self.blobs[shell].shape)
            group['blobs_{}'.format(shell)][...] = self.blobs[shell]

        if isinstance(self.bounds[shell], NautilusBound):
            self.bounds[shell].update(fstream['bound_{}'.format(shell)])

        rng_state = self.rng.bit_generator.state
        group.attrs['rng_state'] = str(rng_state['state']['state'])
        group.attrs['rng_inc'] = str(rng_state['state']['inc'])
        group.attrs['rng_has_uint32'] = rng_state['has_uint32']
        group.attrs['rng_uinteger'] = rng_state['uinteger']

        fstream.close()

    def write_shell_information_update(self, filepath):
        """Update the shell summary statistics.

        Parameters
        ----------
        filepath : string or pathlib.Path
            Path to the file. Must have a '.h5' or '.hdf5' extension.

        """
        with h5py.File(Path(filepath), 'r+') as fstream:
            group = fstream['sampler']
            for key in ['_discard_exploration', 'shell_n', 'shell_n_sample',
                        'shell_n_eff', 'shell_log_l_min', 'shell_log_l',
                        'shell_log_v', 'shell_n_sample_exp', 'shell_end_exp']:
                group.attrs[key] = getattr(self, key)



import numpy as np
import pytest

from scipy.stats import norm

from nautilus.prior import Prior


def test_add_parameter():
    # Test that adding parameters works as expected.

    prior = Prior()
    with pytest.raises(TypeError):
        prior.add_parameter(1.0)

    prior = Prior()
    prior.add_parameter('a')
    prior.add_parameter('b')
    with pytest.raises(ValueError):
        prior.add_parameter('a')

    prior = Prior()
    with pytest.raises(ValueError):
        prior.add_parameter('a', dist='a')

    prior = Prior()
    with pytest.raises(TypeError):
        prior.add_parameter(dist=[0.0])


def test_dimensionality():
    # Test that the dimensionality is correctly reported.

    prior = Prior()
    n_dim = 6
    for i in range(n_dim):
        prior.add_parameter()
    assert prior.dimensionality() == n_dim

    prior = Prior()
    prior.add_parameter()
    prior.add_parameter(dist=5.0)
    prior.add_parameter()
    prior.add_parameter(dist=1)
    assert prior.dimensionality() == 2

    prior = Prior()
    prior.add_parameter(key='a')
    prior.add_parameter(key='b', dist=1)
    prior.add_parameter(key='c', dist='a')
    prior.add_parameter(key='d', dist='b')
    prior.add_parameter(key='e')
    assert prior.dimensionality() == 2


def test_unit_to_physical():
    # Test converting unit hypercube coordinates to physical ones.

    prior = Prior()
    prior.add_parameter(key='a', dist=(-1, +1))
    prior.add_parameter(key='b', dist='a')
    prior.add_parameter(key='c', dist='b')
    dist1 = norm(loc=3.0, scale=2.0)
    dist2 = norm(loc=5.0, scale=1.0)
    prior.add_parameter(key='d', dist=dist1)
    prior.add_parameter(key='e', dist=dist2)

    unit = np.random.random(size=4)
    with pytest.raises(ValueError):
        phys = prior.unit_to_physical(unit)

    unit = np.random.random(size=3)
    phys = prior.unit_to_physical(unit)
    assert unit.shape == phys.shape
    assert np.isclose(phys[0], unit[0] * 2 - 1)
    assert np.isclose(phys[1], dist1.isf(1 - unit[1]))
    assert np.isclose(phys[2], dist2.isf(1 - unit[2]))

    unit = np.random.random(size=(10, 3))
    phys = prior.unit_to_physical(unit)
    assert unit.shape == phys.shape
    assert np.allclose(phys[:, 0], unit[:, 0] * 2 - 1)
    assert np.allclose(phys[:, 1], dist1.isf(1 - unit[:, 1]))
    assert np.allclose(phys[:, 2], dist2.isf(1 - unit[:, 2]))


def test_physical_to_dictionary():
    # Test converting physical coordinates to dictionaries.

    prior = Prior()
    prior.add_parameter(key='a')
    prior.add_parameter(key='b', dist='a')
    prior.add_parameter(key='c', dist='b')
    prior.add_parameter(key='d')
    prior.add_parameter(key='e')
    prior.add_parameter(key='f', dist=0.5)

    phys = np.random.random(size=4)
    with pytest.raises(ValueError):
        param_dict = prior.physical_to_dictionary(phys)

    phys = np.random.random(size=3)
    param_dict = prior.physical_to_dictionary(phys)
    assert isinstance(param_dict, dict)
    assert len(param_dict) == 6
    assert param_dict['a'] == phys[0]
    assert param_dict['d'] == phys[1]
    assert param_dict['e'] == phys[2]
    assert param_dict['b'] == param_dict['a']
    assert param_dict['c'] == param_dict['b']
    assert param_dict['f'] == 0.5

    phys = np.random.random(size=(10, 3))
    param_dict = prior.physical_to_dictionary(phys)
    assert isinstance(param_dict, dict)
    assert len(param_dict.keys()) == 6
    assert np.all(param_dict['a'] == phys[:, 0])
    assert np.all(param_dict['d'] == phys[:, 1])
    assert np.all(param_dict['e'] == phys[:, 2])
    assert np.all(param_dict['b'] == param_dict['a'])
    assert np.all(param_dict['c'] == param_dict['b'])
    assert np.all(param_dict['f'] == 0.5)



"""Neural Network-Boosted Importance Sampling for Bayesian Statistics"""

from .sampler import Sampler
from .prior import Prior

__author__ = 'Johannes U. Lange'
__email__ = 'julange.astro@pm.me'
__version__ = '0.7.0'



import h5py
import numpy as np
import pytest

from pathlib import Path

from nautilus import Sampler
from nautilus.bounds import (
    UnitCube, Ellipsoid, Union, UnitCubeEllipsoidMixture, NeuralBound,
    NautilusBound)
from nautilus.neural import NeuralNetworkEmulator


@pytest.fixture
def h5py_group():
    with h5py.File('test.hdf5', 'w') as filepath:
        yield filepath.create_group('test')
    Path('test.hdf5').unlink()


def test_neural_io(h5py_group):
    # Test that we can write and read a neural network emulator correctly.

    points = np.random.random((100, 2))
    log_l = -np.linalg.norm(points - 0.5, axis=1)
    emulator_write = NeuralNetworkEmulator.train(points, log_l, n_networks=1,
                                                 pool=None)
    emulator_write.write(h5py_group)
    emulator_read = NeuralNetworkEmulator.read(h5py_group)
    assert np.all(emulator_write.predict(points) ==
                  emulator_read.predict(points))


@pytest.mark.parametrize("bound_class", [UnitCube, Ellipsoid, Union,
                                         UnitCubeEllipsoidMixture, NeuralBound,
                                         NautilusBound])
@pytest.mark.parametrize("rng_sync", [True, False])
def test_bounds_io(h5py_group, bound_class, rng_sync):
    # Test that we can write and read a bound correctly. In particular, also
    # test that the random number generator is correctly set after writing and
    # reading.

    n_dim = 5
    n_points = 100
    np.random.seed(0)
    points = np.random.random((n_points, n_dim))
    rng = np.random.default_rng(0)

    if bound_class == UnitCube:
        args = (n_dim, )
        kwargs = dict()
    elif bound_class in [Ellipsoid, Union, UnitCubeEllipsoidMixture]:
        args = (points, )
        kwargs = dict()
    else:
        log_l = -np.linalg.norm(points - 0.5, axis=1)
        log_l_min = np.median(log_l)
        if bound_class == NeuralBound:
            args = (points, log_l, log_l_min)
        else:
            args = (points, log_l, log_l_min, np.log(0.5))
        kwargs = dict(n_networks=1, pool=None)

    bound_write = bound_class.compute(*args, **kwargs, rng=rng)
    if bound_class == Union:
        bound_write.split_bound()

    bound_write.write(h5py_group)
    rng = np.random.default_rng(1)
    if rng_sync:
        if bound_class == NeuralBound:
            rng = None
        elif bound_class == UnitCubeEllipsoidMixture:
            rng.bit_generator.state = bound_write.cube.rng.bit_generator.state
        else:
            rng.bit_generator.state = bound_write.rng.bit_generator.state

    bound_read = bound_class.read(h5py_group, rng=rng)

    if bound_class != NeuralBound:
        assert (np.all(bound_write.sample(10000) == bound_read.sample(10000))
                == rng_sync)

    if ((rng_sync or bound_class in [UnitCube, Ellipsoid]) and not
            bound_class == NeuralBound):
        assert bound_write.volume() == bound_read.volume()

    points = np.random.random((10000, n_dim))
    assert np.all(bound_write.contains(points) == bound_read.contains(points))

    if bound_class in [Union, NautilusBound]:
        bound_write.sample(10000)
        bound_write.update(h5py_group)

        rng = np.random.default_rng(1)
        if rng_sync:
            rng.bit_generator.state = bound_write.rng.bit_generator.state

        bound_read = bound_class.read(h5py_group, rng=rng)

        assert (np.all(bound_write.sample(20000) == bound_read.sample(20000))
                == rng_sync)

        if rng_sync:
            assert bound_write.volume() == bound_read.volume()


@pytest.mark.parametrize("blobs", [True, False])
@pytest.mark.parametrize("discard_exploration", [True, False])
@pytest.mark.parametrize("n_networks", [0, 1, 2])
def test_sampler_io(blobs, discard_exploration, n_networks):
    # Test that we can write and read a sampler correctly. In particular, also
    # test that the random number generator is correctly set after writing and
    # reading. Also make sure that the sampler can print out the progress.

    def prior(x):
        return x

    def likelihood(x):
        if blobs:
            return -np.linalg.norm(x - 0.5) * 0.001, x[0]
        else:
            return -np.linalg.norm(x - 0.5) * 0.001

    sampler_write = Sampler(prior, likelihood, n_dim=2, n_live=100,
                            n_networks=n_networks, filepath='test.hdf5',
                            resume=False, seed=0)
    sampler_write.run(f_live=0.45, n_eff=1000,
                      discard_exploration=discard_exploration, verbose=True)
    sampler_write.explored = False
    # Reset the interal shell calculations.
    sampler_write.discard_exploration = discard_exploration
    sampler_read = Sampler(prior, likelihood, n_dim=2, n_live=100,
                           n_networks=n_networks, filepath='test.hdf5',
                           resume=True)
    sampler_read.explored = False
    # Reset the interal shell calculations.
    sampler_read.discard_exploration = discard_exploration
    assert sampler_write.evidence() == sampler_read.evidence()

    sampler_write.run(f_live=0.45, n_eff=5000,
                      discard_exploration=discard_exploration, verbose=True)
    sampler_read.run(f_live=0.45, n_eff=5000,
                     discard_exploration=discard_exploration, verbose=True)

    posterior_write = sampler_write.posterior()
    posterior_read = sampler_read.posterior()

    for arr_write, arr_read in zip(posterior_write, posterior_read):
        assert np.all(arr_write == arr_read)

    assert sampler_write.evidence() == sampler_read.evidence()

    Path('test.hdf5').unlink()



import numpy as np
import pytest

from nautilus import Sampler


@pytest.fixture
def prior():
    def prior(x):
        return x
    return prior


@pytest.mark.parametrize("dtype", [np.float64, np.int64])
@pytest.mark.parametrize("vectorized", [True, False])
@pytest.mark.parametrize("discard_exploration", [True, False])
def test_blobs_single(prior, dtype, vectorized, discard_exploration):
    # Test that blobs work and their data types and values match.

    def likelihood(x):
        if vectorized:
            return np.ones(len(x)), (10 * x[:, 0]).astype(dtype)
        else:
            return 1, dtype(10 * x[0])

    sampler = Sampler(prior, likelihood, n_dim=2, n_live=10,
                      vectorized=vectorized)
    sampler.run(f_live=1.0, n_eff=200, discard_exploration=discard_exploration)
    points, log_w, log_l, blobs = sampler.posterior(return_blobs=True)

    assert len(points) == len(blobs)
    assert blobs.dtype == dtype
    assert np.all((10 * points[:, 0]).astype(dtype) == blobs)


@pytest.mark.parametrize("vectorized", [True, False])
@pytest.mark.parametrize("discard_exploration", [True, False])
def test_blobs_multi(prior, vectorized, discard_exploration):
    # Test that blobs work and their data types and values match.

    def likelihood(x):
        if vectorized:
            return (np.ones(len(x)), x[:, 0].astype(np.float32),
                    x[:, 1].astype(np.float32))
        else:
            return 1, np.float32(x[0]), np.float32(x[1])

    sampler = Sampler(prior, likelihood, n_dim=2, n_live=10,
                      vectorized=vectorized)
    sampler.run(f_live=1.0, n_eff=200, discard_exploration=discard_exploration)
    points, log_w, log_l, blobs = sampler.posterior(return_blobs=True)

    assert len(points) == len(blobs)
    assert blobs['blob_0'].dtype == np.float32
    assert blobs['blob_1'].dtype == np.float32
    assert np.all(points[:, 0].astype(np.float32) == blobs['blob_0'])
    assert np.all(points[:, 1].astype(np.float32) == blobs['blob_1'])


@pytest.mark.parametrize("vectorized", [True, False])
@pytest.mark.parametrize("discard_exploration", [True, False])
def test_blobs_dtype(prior, vectorized, discard_exploration):
    # Test that blobs work and their data types and values match.

    def likelihood(x):
        if vectorized:
            return np.ones(len(x)), x[:, 0], x[:, 1]
        else:
            return 1, x[0], x[1]

    blobs_dtype = [('a', '|S10'), ('b', np.int16)]
    sampler = Sampler(prior, likelihood, n_dim=2, n_live=10,
                      vectorized=vectorized, blobs_dtype=blobs_dtype)
    sampler.run(f_live=1.0, n_eff=200, discard_exploration=discard_exploration)
    points, log_w, log_l, blobs = sampler.posterior(return_blobs=True)

    assert len(points) == len(blobs)
    assert blobs['a'].dtype == blobs_dtype[0][1]
    assert blobs['b'].dtype == blobs_dtype[1][1]
    assert np.all(points[:, 0].astype(blobs_dtype[0][1]) == blobs['a'])
    assert np.all(points[:, 1].astype(blobs_dtype[1][1]) == blobs['b'])


@pytest.mark.parametrize("vectorized", [True, False])
@pytest.mark.parametrize("discard_exploration", [True, False])
def test_blobs_single_dtype(prior, vectorized, discard_exploration):
    # Test that blobs work even when returning specyfing a single dtype like
    # CosmoSIS does.

    def likelihood(x):
        if vectorized:
            return np.ones(len(x)), x[:, 0], x[:, 1]
        else:
            return 1, x[0], x[1]

    sampler = Sampler(prior, likelihood, n_dim=2, n_live=10,
                      vectorized=vectorized, blobs_dtype=float)
    sampler.run(f_live=1.0, n_eff=200, discard_exploration=discard_exploration)
    points, log_w, log_l, blobs = sampler.posterior(return_blobs=True)
    assert len(points) == len(blobs)
    assert np.all(points[:, 0] == blobs[:, 0])
    assert np.all(points[:, 1] == blobs[:, 1])



"""Module implementing basic multi-dimensional bounds."""

import numpy as np
from scipy.special import gammaln
from scipy.linalg.lapack import dpotrf, dpotri
from threadpoolctl import threadpool_limits


class UnitCube():
    r"""Unit (hyper)cube bound.

    The :math:`n`-dimensional unit hypercube has :math:`n_{\rm dim}` parameters
    :math:`x_i` with :math:`0 \leq x_i < 1` for all :math:`x_i`.

    Attributes
    ----------
    n_dim : int
        Number of dimensions.
    rng : numpy.random.Generator
        Determines random number generation.

    """

    @classmethod
    def compute(cls, n_dim, rng=None):
        """Compute the bound.

        Parameters
        ----------
        n_dim : int
            Number of dimensions.
        rng : None or numpy.random.Generator, optional
            Determines random number generation. Default is None.

        Returns
        -------
        bound : UnitCube
            The bound.

        """
        bound = cls()
        bound.n_dim = n_dim

        if rng is None:
            bound.rng = np.random.default_rng()
        else:
            bound.rng = rng

        return bound

    def contains(self, points):
        """Check whether points are contained in the bound.

        Parameters
        ----------
        points : numpy.ndarray
            A 1-D or 2-D array containing single point or a collection of
            points. If more than one-dimensional, each row represents a point.

        Returns
        -------
        in_bound : bool or numpy.ndarray
            Bool or array of bools describing for each point whether it is
            contained in the bound.

        """
        return np.all((points >= 0) & (points < 1), axis=-1)

    def sample(self, n_points=100, pool=None):
        """Sample points from the bound.

        Parameters
        ----------
        n_points : int, optional
            How many points to draw. Default is 100.
        pool : ignored
            Not used. Present for API consistency.

        Returns
        -------
        points : numpy.ndarray
            Points as two-dimensional array of shape (n_points, n_dim).

        """
        points = self.rng.random(size=(n_points, self.n_dim))
        return points

    def volume(self):
        """Return the natural log of the volume of the bound.

        Returns
        -------
        float
            The natural log of the volume of the bound.

        """
        return 0

    def write(self, group):
        """Write the bound to an HDF5 group.

        Parameters
        ----------
        group : h5py.Group
            HDF5 group to write to.

        """
        group.attrs['type'] = 'UnitCube'
        group.attrs['n_dim'] = self.n_dim

    @classmethod
    def read(cls, group, rng=None):
        """Read the bound from an HDF5 group.

        Parameters
        ----------
        group : h5py.Group
            HDF5 group to write to.
        rng : None or numpy.random.Generator, optional
            Determines random number generation. Default is None.

        Returns
        -------
        bound : UnitCube
            The bound.

        """
        bound = cls()

        if rng is None:
            bound.rng = np.random.default_rng()
        else:
            bound.rng = rng

        bound.n_dim = group.attrs['n_dim']

        return bound

    def reset(self, rng=None):
        """Reset random number generation and any progress, if applicable.

        Parameters
        ----------
        rng : None or numpy.random.Generator, optional
            Determines random number generation. If None, random number
            generation is not reset. Default is None.

        """
        if rng is not None:
            self.rng = rng


def invert_symmetric_positive_semidefinite_matrix(m):
    """Invert a symmetric positive sem-definite matrix.

    This function is faster than numpy.linalg.inv but does not work for
    arbitrary matrices.

    Parameters
    ----------
    m : numpy.ndarray
        Matrix to be inverted.

    Returns
    -------
    m_inv : numpy.ndarray
        Inverse of the matrix.

    """
    m_inv_triangle = dpotri(dpotrf(m)[0])[0]
    return m_inv_triangle + m_inv_triangle.T - np.diag(np.diag(m_inv_triangle))


def minimum_volume_enclosing_ellipsoid(points, tol=0, max_iterations=1000):
    r"""Find an approximation to the minimum volume enclosing ellipsoid (MVEE).

    This functions finds an approximation to the MVEE using the Khachiyan
    algorithm.

    This function is based on
    http://citeseerx.ist.psu.edu/viewdoc/summary?doi=10.1.1.116.7691 but
    implemented independently of the corresponding MATLAP implementation or
    other Python ports of the MATLAP code.

    Parameters
    ----------
    points : numpy.ndarray with shape (n_points, n_dim)
        A 2-D array where each row represents a point.
    tol : float, optional
        Tolerance parameter for early stopping. Smaller value lead to more
        accurate results. Default is 0.
    max_iterations : int, optional
        Maximum number of iterations before the function is stopped. Default
        is 1000.

    Returns
    -------
    c : numpy.ndarray
        The position of the ellipsoid.
    A : numpy.ndarray
        The bounds of the ellipsoid in matrix form, i.e.
        :math:`(x - c)^T A (x - c) \leq 1`.

    """
    m, n = points.shape
    q = np.append(points, np.ones(shape=(m, 1)), axis=1)
    u = np.repeat(1.0 / m, m)
    q_outer = np.array([np.outer(q_i, q_i) for q_i in q])
    e = np.diag(np.ones(m))

    for i in range(max_iterations):
        if i % 1000 == 0:
            v = np.einsum('ji,j,jk', q, u, q)
        g = np.einsum('ijk,jk', q_outer,
                      invert_symmetric_positive_semidefinite_matrix(v))
        j = np.argmax(g)
        d_u = e[j] - u
        a = (g[j] - (n + 1)) / ((n + 1) * (g[j] - 1))
        shift = np.linalg.norm(a * d_u)
        v = v * (1 - a) + a * q_outer[j]
        u = u + a * d_u
        if shift <= tol:
            break

    c = np.einsum('i,ij', u, points)
    A_inv = (np.einsum('ji,j,jk', points, u, points) - np.outer(c, c)) * n
    A = np.linalg.inv(A_inv)

    scale = np.amax(np.einsum('...i,ij,...j', points - c, A, points - c))
    A /= scale

    return c, A


class Ellipsoid():
    r"""Ellipsoid bound.

    Attributes
    ----------
    n_dim : int
        Number of dimensions.
    c : numpy.ndarray
        The position of the ellipsoid.
    A : numpy.ndarray
        The bounds of the ellipsoid in matrix form, i.e.
        :math:`(x - c)^T A (x - c) \leq 1`.
    B : numpy.ndarray
        Cholesky decomposition of the inverse of A.
    B_inv : numpy.ndarray
        Inverse of B.
    log_v : float
        Natural log of the volume of the ellipsoid.
    rng : numpy.random.Generator
        Determines random number generation.

    """

    @classmethod
    def compute(cls, points, enlarge_per_dim=1.1, max_iterations=1000,
                rng=None):
        """Compute the bound.

        Parameters
        ----------
        points : numpy.ndarray with shape (n_points, n_dim)
            A 2-D array where each row represents a point.
        enlarge_per_dim : float, optional
            Along each dimension, the ellipsoid is enlarged by this factor.
            Default is 1.1.
        max_iterations : int, optional
            Maximum number of iterations before the minimization algorithm for
            the minimum volume enclosing ellipsoid is stopped. Ignored if
            `fast` is True. Default is 1000.
        rng : None or numpy.random.Generator, optional
            Determines random number generation. Default is None.

        Raises
        ------
        ValueError
            If `enlarge_per_dim` is smaller than unity or the number of points
            does not exceed the number of dimensions.

        Returns
        -------
        bound : Ellipsoid
            The bound.

        """
        bound = cls()
        bound.n_dim = points.shape[1]

        if enlarge_per_dim < 1.0:
            raise ValueError(
                "The 'enlarge_per_dim' factor cannot be smaller than unity.")

        if not points.shape[0] > bound.n_dim:
            raise ValueError('Number of points must be larger than number ' +
                             'dimensions.')

        with threadpool_limits(limits=1):
            bound.c, bound.A = minimum_volume_enclosing_ellipsoid(
                points, max_iterations=max_iterations)

        bound.A /= enlarge_per_dim**2.0
        bound.B = np.linalg.cholesky(np.linalg.inv(bound.A))
        bound.B_inv = np.linalg.inv(bound.B)
        bound.log_v = (np.linalg.slogdet(bound.B)[1] +
                       bound.n_dim * np.log(2.) +
                       bound.n_dim * gammaln(1.5) -
                       gammaln(bound.n_dim / 2.0 + 1))

        if rng is None:
            bound.rng = np.random.default_rng()
        else:
            bound.rng = rng

        return bound

    def transform(self, points, inverse=False):
        """Transform points into the coordinate frame of the ellipsoid.

        Parameters
        ----------
        points : numpy.ndarray
            A 1-D or 2-D array containing single point or a collection of
            points. If more than one-dimensional, each row represents a point.
        inverse : bool, optional
            By default, the coordinates are transformed from the regular
            coordinates to the coordinates in the ellipsoid. If `inverse` is
            set to true, this function does the inverse operation, i.e.
            transform from the ellipsoidal to the regular coordinates. Default
            is False.

        Returns
        -------
        points_t : numpy.ndarray
            Transformed points.

        """
        if not inverse:
            return np.einsum('ij, ...j', self.B_inv, points - self.c)
        else:
            return np.einsum('ij, ...j', self.B, points) + self.c

    def contains(self, points):
        """Check whether points are contained in the bound.

        Parameters
        ----------
        points : numpy.ndarray
            A 1-D or 2-D array containing single point or a collection of
            points. If more than one-dimensional, each row represents a point.

        Returns
        -------
        in_bound : bool or numpy.ndarray
            Bool or array of bools describing for each point whether it is
            contained in the bound.

        """
        return np.sum(self.transform(points)**2, axis=-1) < 1

    def sample(self, n_points=100):
        """Sample points from the bound.

        Parameters
        ----------
        n_points : int, optional
            How many points to draw. Default is 100.

        Returns
        -------
        points : numpy.ndarray
            Points as two-dimensional array of shape (n_points, n_dim).

        """
        points = self.rng.normal(size=(n_points, self.n_dim))
        points = points / np.sqrt(np.sum(points**2, axis=1))[:, np.newaxis]
        points *= self.rng.uniform(size=n_points)[:, np.newaxis]**(
            1.0 / self.n_dim)
        points = self.transform(points, inverse=True)
        return points

    def volume(self):
        """Return the natural log of the volume of the bound.

        Returns
        -------
        log_v : float
            The natural log of the volume.

        """
        return self.log_v

    def write(self, group):
        """Write the bound to an HDF5 group.

        Parameters
        ----------
        group: h5py.Group
            HDF5 group to write to.

        """
        group.attrs['type'] = 'Ellipsoid'
        for key in ['n_dim', 'c', 'A', 'B', 'B_inv', 'log_v']:
            group.attrs[key] = getattr(self, key)

    @classmethod
    def read(cls, group, rng=None):
        """Read the bound from an HDF5 group.

        Parameters
        ----------
        group: h5py.Group
            HDF5 group to write to.
        rng: None or numpy.random.Generator, optional
            Determines random number generation. Default is None.

        Returns
        -------
        bound: Ellipsoid
            The bound.

        """
        bound = cls()

        if rng is None:
            bound.rng = np.random.default_rng()
        else:
            bound.rng = rng

        for key in ['n_dim', 'c', 'A', 'B', 'B_inv', 'log_v']:
            setattr(bound, key, group.attrs[key])

        return bound

    def reset(self, rng=None):
        """Reset random number generation and any progress, if applicable.

        Parameters
        ----------
        rng : None or numpy.random.Generator, optional
            Determines random number generation. If None, random number
            generation is not reset. Default is None.

        """
        if rng is not None:
            self.rng = rng


class UnitCubeEllipsoidMixture():
    """Mixture of a unit cube and an ellipsoid.

    Dimensions along which an ellipsoid has a smaller volume than a unit cube
    are defined via ellipsoids and vice versa.

    Attributes
    ----------
    n_dim: int
        Number of dimensions.
    dim_cube: numpy.ndarray
        Whether the boundary in each dimension is defined via the unit cube.
    cube: UnitCube or None
        Unit cube defining the boundary along certain dimensions.
    ellipsoid: Ellipsoid or None
        Ellipsoid defining the boundary along certain dimensions.

    """

    @classmethod
    def compute(cls, points, enlarge_per_dim=1.1, rng=None):
        """Compute the bound.

        Parameters
        ----------
        points: numpy.ndarray with shape(n_points, n_dim)
            A 2-D array where each row represents a point.
        enlarge_per_dim: float, optional
            Along each dimension, the ellipsoid is enlarged by this factor.
            Default is 1.1.
        rng: None or numpy.random.Generator, optional
            Determines random number generation. Default is None.

        Returns
        -------
        bound: UnitCubeEllipsoidMixture
            The bound.

        """
        bound = cls()
        bound.n_dim = points.shape[1]

        kwargs = dict(enlarge_per_dim=enlarge_per_dim, max_iterations=1000,
                      rng=rng)

        # First, start by sampling all dimensions using an ellipsoid..
        ellipsoid = Ellipsoid.compute(points, **kwargs)
        bound.dim_cube = np.zeros(bound.n_dim, dtype=bool)

        # Sample dimensions from the unit cube until volume doesn't decrease.
        while np.sum(~bound.dim_cube) > 1:
            c = ellipsoid.c
            A_inv = np.linalg.inv(ellipsoid.A)

            # Start by projecting the current ellipsoid along each dimension to
            # estimate dropping which dimension is most likely to lead to the
            # biggest volume reduction.
            log_v = np.zeros(np.sum(~bound.dim_cube))
            for i in range(np.sum(~bound.dim_cube)):
                points_proj = np.delete(points[:, ~bound.dim_cube], i, axis=1)
                c_proj = np.delete(c, i)
                A_inv_proj = np.delete(np.delete(A_inv, i, axis=0), i, axis=1)
                A_proj = np.linalg.inv(A_inv_proj)
                scale = np.amax(np.einsum('...i,ij,...j', points_proj - c_proj,
                                A_proj, points_proj - c_proj))
                A_proj /= scale
                log_v[i] = np.linalg.slogdet(np.linalg.inv(A_proj))[1]

            dim = np.arange(bound.n_dim)[~bound.dim_cube][np.argmin(log_v)]
            bound.dim_cube[dim] = True
            ellipsoid_proj = Ellipsoid.compute(
                points[:, ~bound.dim_cube], **kwargs)

            if ellipsoid_proj.volume() < ellipsoid.volume():
                ellipsoid = ellipsoid_proj
            else:
                bound.dim_cube[dim] = False
                break

        # The above algorithm will not necessarily find the optimal combination
        # of which dimensions to sample from the unit cube. In particular, the
        # ellipsoid may have a volume larger than the unit cube. If that's the
        # case, start from the unit cube and add dimensions to sample from
        # ellipsoids until volume doesn't decrease.
        if ellipsoid.volume() > 0:
            ellipsoid = UnitCube.compute(points)
            bound.dim_cube = np.ones(bound.n_dim, dtype=bool)
            # Check which dimensions are better sampled from an ellipsoid,
            # i.e., have smaller volumes.
            tested = np.zeros(bound.n_dim, dtype=bool)
            while ~np.all(tested):
                for dim in np.arange(bound.n_dim)[~tested]:
                    bound.dim_cube[dim] = False
                    tested[dim] = True
                    ellipsoid_test = Ellipsoid.compute(
                        points[:, ~bound.dim_cube], **kwargs)
                    if ellipsoid.volume() > ellipsoid_test.volume():
                        ellipsoid = ellipsoid_test
                        tested[bound.dim_cube] = False
                    else:
                        bound.dim_cube[dim] = True

        if np.any(bound.dim_cube):
            bound.cube = UnitCube.compute(np.sum(bound.dim_cube), rng=rng)
        else:
            bound.cube = None

        if np.all(bound.dim_cube):
            bound.ellipsoid = None
        else:
            bound.ellipsoid = ellipsoid

        return bound

    def transform(self, points):
        """Transform points into the frame of the cube-ellipsoid mixture.

        Along dimensions where the boundary is not defined via the unit range,
        the coordinates are transformed into the range [-1, +1]. Along all
        other dimensions, they are transformed into the coordinate system of
        the bounding ellipsoid.

        Parameters
        ----------
        points: numpy.ndarray
            A 1-D or 2-D array containing single point or a collection of
            points. If more than one-dimensional, each row represents a point.

        Returns
        -------
        points_t: numpy.ndarray
            Transformed points.

        """
        points_t = np.copy(points)
        if self.cube is not None:
            idx = np.arange(self.n_dim)[self.dim_cube]
            points_t[:, idx] = points[:, idx] * 2 - 1
        if self.ellipsoid is not None:
            idx = np.arange(self.n_dim)[~self.dim_cube]
            points_t[:, idx] = self.ellipsoid.transform(points[:, idx])
        return points_t

    def contains(self, points):
        """Check whether points are contained in the bound.

        Parameters
        ----------
        points: numpy.ndarray
            A 1-D or 2-D array containing single point or a collection of
            points. If more than one-dimensional, each row represents a point.

        Returns
        -------
        in_bound: bool or numpy.ndarray
            Bool or array of bools describing for each point whether it is
            contained in the bound.

        """
        in_bound = np.ones(points.shape[:-1], dtype=bool)
        if self.cube is not None:
            idx = np.arange(self.n_dim)[self.dim_cube]
            in_bound = in_bound & self.cube.contains(points[..., idx])
        if self.ellipsoid is not None:
            idx = np.arange(self.n_dim)[~self.dim_cube]
            in_bound = in_bound & self.ellipsoid.contains(points[..., idx])
        return in_bound

    def sample(self, n_points=100):
        """Sample points from the bound.

        Parameters
        ----------
        n_points: int, optional
            How many points to draw. Default is 100.

        Returns
        -------
        points: numpy.ndarray
            Points as two-dimensional array of shape (n_points, n_dim).

        """
        points = np.zeros((n_points, self.n_dim))
        if self.cube is not None:
            idx = np.arange(self.n_dim)[self.dim_cube]
            points[:, idx] = self.cube.sample(n_points)
        if self.ellipsoid is not None:
            idx = np.arange(self.n_dim)[~self.dim_cube]
            points[:, idx] = self.ellipsoid.sample(n_points)
        return points

    def volume(self):
        """Return the natural log of the volume of the bound.

        Returns
        -------
        log_v: float
            The natural log of the volume.

        """
        if self.ellipsoid is None:
            return 0
        else:
            return self.ellipsoid.volume()

    def write(self, group):
        """Write the bound to an HDF5 group.

        Parameters
        ----------
        group: h5py.Group
            HDF5 group to write to.

        """
        group.attrs['type'] = 'UnitCubeEllipsoidMixture'
        group.attrs['n_dim'] = self.n_dim

        group.create_dataset('dim_cube', data=self.dim_cube)
        if self.cube is not None:
            self.cube.write(group.create_group('cube'))
        if self.ellipsoid is not None:
            self.ellipsoid.write(group.create_group('ellipsoid'))

    @classmethod
    def read(cls, group, rng=None):
        """Read the bound from an HDF5 group.

        Parameters
        ----------
        group: h5py.Group
            HDF5 group to write to.
        rng: None or numpy.random.Generator, optional
            Determines random number generation. Default is None.

        Returns
        -------
        bound: UnitCubeEllipsoidMixture
            The bound.

        """
        bound = cls()

        if rng is None:
            rng = np.random.default_rng()

        bound.n_dim = group.attrs['n_dim']
        bound.dim_cube = np.array(group['dim_cube'])

        if np.any(bound.dim_cube):
            bound.cube = UnitCube.read(group['cube'], rng=rng)
        else:
            bound.cube = None

        if not np.all(bound.dim_cube):
            bound.ellipsoid = Ellipsoid.read(group['ellipsoid'], rng=rng)
        else:
            bound.ellipsoid = None

        return bound

    def reset(self, rng=None):
        """Reset random number generation and any progress, if applicable.

        Parameters
        ----------
        rng : None or numpy.random.Generator, optional
            Determines random number generation. If None, random number
            generation is not reset. Default is None.

        """
        if rng is not None:
            if self.ellipsoid is not None:
                self.ellipsoid.reset(rng)
            if self.cube is not None:
                self.cube.reset(rng)



"""Module implementing multi-dimensional neural network-based bounds."""

import numpy as np
from functools import partial
from scipy.stats import percentileofscore
from threadpoolctl import threadpool_limits

from .basic import Ellipsoid, UnitCubeEllipsoidMixture
from .union import Union
from ..neural import NeuralNetworkEmulator


class NeuralBound():
    """Neural network-based bound.

    Attributes
    ----------
    n_dim : int
        Number of dimensions.
    outer_bound : UnitCubeEllipsoidMixture
        Outer bound around the points above the likelihood threshold.
    emulator : object
        Emulator based on `sklearn.neural_network.MLPRegressor` used to fit and
        predict likelihood scores.
    score_predict_min : float
        Minimum score predicted by the emulator to be considered part of the
        bound.
    """

    @classmethod
    def compute(cls, points, log_l, log_l_min, enlarge_per_dim=1.1,
                n_networks=4, neural_network_kwargs={}, pool=None,
                rng=None):
        """Compute a neural network-based bound.

        Parameters
        ----------
        points : numpy.ndarray with shape (m, n)
            A 2-D array where each row represents a point.
        log_l : numpy.ndarray of length m
            Likelihood of each point.
        log_l_min : float
            Target likelihood threshold of the bound.
        enlarge_per_dim : float, optional
            Along each dimension, the ellipsoid of the outer bound is enlarged
            by this factor. Default is 1.1.
        n_networks : int, optional
            Number of networks used in the emulator. Default is 4.
        neural_network_kwargs : dict, optional
            Non-default keyword arguments passed to the constructor of
            MLPRegressor.
        pool : multiprocessing.Pool, optional
            Pool used for parallel processing.
        rng : None or numpy.random.Generator, optional
            Determines random number generation. Default is None.

        Returns
        -------
        bound : NeuralBound
            The bound.

        """
        bound = cls()
        bound.n_dim = points.shape[1]

        if rng is None:
            rng = np.random.default_rng()

        # Determine the outer bound.
        bound.outer_bound = Ellipsoid.compute(
            points[log_l >= log_l_min], enlarge_per_dim=enlarge_per_dim,
            rng=rng)

        if n_networks == 0:
            bound.emulator = None
            bound.score_predict_min = 0
            return bound

        # Train the network.
        select = bound.outer_bound.contains(points)
        points = points[select]
        log_l = log_l[select]

        points_t = bound.outer_bound.transform(points)
        perc = np.argsort(np.argsort(log_l)) / float(len(log_l))
        perc_min = percentileofscore(log_l, log_l_min) / 100
        score = np.zeros(len(points))
        select = perc < perc_min
        if np.any(select):
            score[select] = 0.5 * (perc[select] / perc_min)
        score[~select] = 1 - 0.5 * (1 - perc[~select]) / (1 - perc_min)
        bound.emulator = NeuralNetworkEmulator.train(
            points_t, score, n_networks=n_networks,
            neural_network_kwargs=neural_network_kwargs, pool=pool)

        bound.score_predict_min = np.polyval(np.polyfit(
            score, bound.emulator.predict(points_t), 3), 0.5)

        return bound

    def contains(self, points):
        """Check whether points are contained in the bound.

        Parameters
        ----------
        points : numpy.ndarray
            A 1-D or 2-D array containing single point or a collection of
            points. If more than one-dimensional, each row represents a point.

        Returns
        -------
        in_bound : bool or numpy.ndarray
            Bool or array of bools describing for each point if it is contained
            in the bound.

        """
        points = np.atleast_2d(points)

        in_bound = self.outer_bound.contains(points)
        if np.any(in_bound) and self.emulator is not None:
            points_t = self.outer_bound.transform(points)
            in_bound[in_bound] = (self.emulator.predict(points_t[in_bound]) >
                                  self.score_predict_min)

        return in_bound

    def write(self, group):
        """Write the bound to an HDF5 group.

        Parameters
        ----------
        group : h5py.Group
            HDF5 group to write to.

        """
        group.attrs['n_dim'] = self.n_dim
        group.attrs['score_predict_min'] = self.score_predict_min
        self.outer_bound.write(group.create_group('outer_bound'))
        if self.emulator is not None:
            self.emulator.write(group.create_group('emulator'))

    @classmethod
    def read(cls, group, rng=None):
        """Read the bound from an HDF5 group.

        Parameters
        ----------
        group : h5py.Group
            HDF5 group to write to.
        rng : None or numpy.random.Generator, optional
            Determines random number generation. Default is None.

        Returns
        -------
        bound : NeuralBound
            The bound.

        """
        bound = cls()

        if rng is None:
            rng = np.random.default_rng()

        bound.n_dim = group.attrs['n_dim']
        bound.score_predict_min = group.attrs['score_predict_min']
        bound.outer_bound = Ellipsoid.read(group['outer_bound'], rng=rng)
        if 'emulator' in group:
            bound.emulator = NeuralNetworkEmulator.read(group['emulator'])
        else:
            bound.emulator = None

        return bound


class NautilusBound():
    """Union of multiple non-overlapping neural network-based bounds.

    The bound is the overlap of the union of multiple neural network-based
    bounds and the unit hypercube.

    Attributes
    ----------
    n_dim : int
        Number of dimensions.
    neural_bounds : list
        List of the individual neural network-based bounds.
    outer_bound : Union
        Outer bound used for sampling.
    rng : None or numpy.random.Generator
        Determines random number generation.
    points : numpy.ndarray
        Points that a call to `sample` will return next.
    n_sample : int
        Number of points sampled from the outer bound.
    n_reject : int
        Number of points rejected due to not falling into the neural
        network-based bounds.
    """

    @classmethod
    def compute(cls, points, log_l, log_l_min, log_v_target,
                enlarge_per_dim=1.1, n_points_min=None, split_threshold=100,
                n_networks=4, neural_network_kwargs={}, pool=None,
                rng=None):
        """Compute a union of multiple neural network-based bounds.

        Parameters
        ----------
        points : numpy.ndarray with shape (m, n)
            A 2-D array where each row represents a point.
        log_l : numpy.ndarray of length m
            Likelihood of each point.
        log_l_min : float
            Target likelihood threshold of the bound.
        log_v_target : float
            Expected volume of the bound. Used for multi-ellipsoidal
            decomposition.
        enlarge_per_dim : float, optional
            Along each dimension, the ellipsoid of the outer bound is enlarged
            by this factor. Default is 1.1.
        n_points_min : int or None, optional
            The minimum number of points each ellipsoid should have.
            Effectively, ellipsoids with less than twice that number will not
            be split further. If None, uses `n_points_min = n_dim + 1`. Default
            is None.
        split_threshold: float, optional
            Threshold used for splitting the multi-ellipsoidal bound used for
            sampling. If the volume of the bound is larger than
            `split_threshold` times the target volume, the multi-ellipsiodal
            bound is split further, if possible. Default is 100.
        n_networks : int, optional
            Number of networks used in the emulator. Default is 4.
        neural_network_kwargs : dict, optional
            Non-default keyword arguments passed to the constructor of
            MLPRegressor.
        pool : multiprocessing.Pool, optional
            Pool used for parallel processing.
        rng : None or numpy.random.Generator, optional
            Determines random number generation. Default is None.

        Returns
        -------
        bound : NautilusBound
            The bound.

        """
        bound = cls()
        bound.n_dim = points.shape[1]

        bound.neural_bounds = []

        multi_ellipsoid = Union.compute(
            points[log_l >= log_l_min], enlarge_per_dim=enlarge_per_dim,
            n_points_min=n_points_min, bound_class=Ellipsoid,
            rng=rng)

        while multi_ellipsoid.split_bound(allow_overlap=False):
            pass

        for ellipsoid in multi_ellipsoid.bounds:
            select = ellipsoid.contains(points)
            bound.neural_bounds.append(NeuralBound.compute(
                points[select], log_l[select], log_l_min,
                enlarge_per_dim=enlarge_per_dim, n_networks=n_networks,
                neural_network_kwargs=neural_network_kwargs, pool=pool,
                rng=rng))

        bound.outer_bound = Union.compute(
            points[log_l >= log_l_min], enlarge_per_dim=enlarge_per_dim,
            n_points_min=n_points_min, bound_class=UnitCubeEllipsoidMixture,
            rng=rng)

        while bound.outer_bound.volume() - log_v_target > np.log(
                split_threshold * enlarge_per_dim**points.shape[1]):
            if not bound.outer_bound.split_bound():
                break

        if rng is None:
            bound.rng = np.random.default_rng()
        else:
            bound.rng = rng

        bound.points = np.zeros((0, points.shape[1]))
        bound.n_sample = 0
        bound.n_reject = 0

        return bound

    def contains(self, points):
        """Check whether points are contained in the bound.

        Parameters
        ----------
        points : numpy.ndarray
            A 1-D or 2-D array containing single point or a collection of
            points. If more than one-dimensional, each row represents a point.

        Returns
        -------
        in_bound : bool or numpy.ndarray
            Bool or array of bools describing for each point it is contained in
            the bound.

        """
        in_bound = self.outer_bound.contains(points)
        if len(self.neural_bounds) > 0:
            in_bound = in_bound & np.any(
                [bound.contains(points) for bound in self.neural_bounds],
                axis=0)
        return in_bound

    @threadpool_limits.wrap(limits=1)
    def _reset_and_sample(self, n_points=100, rng=None):
        """Reset the bound, sample points internally and return the result.

        Parameters
        ----------
        n_points : int, optional
            How many points to draw. Default is 100.
        rng : None or numpy.random.Generator, optional
            Determines random number generation. If None, random number
            generation is not reset. Default is None.

        Returns
        -------
        bound : NautilusBound
            The bound.

        """
        self.reset(rng=rng)
        self.sample(n_points=n_points, return_points=False)
        return self

    def sample(self, n_points=100, return_points=True, pool=None):
        """Sample points from the the bound.

        Parameters
        ----------
        n_points : int, optional
            How many points to draw. Default is 100.
        return_points : bool, optional
            If True, return sampled points. Otherwise, sample internally until
            at least `n_points` are saved.
        pool : multiprocessing.Pool, optional
            Pool used for parallel processing.

        Returns
        -------
        points : numpy.ndarray
            Points as two-dimensional array of shape (n_points, n_dim).

        """
        if len(self.points) < n_points:
            if pool is None:
                while len(self.points) < n_points:
                    n_sample = 1000
                    points = self.outer_bound.sample(n_sample)
                    in_bound = np.any([bound.contains(points) for bound in
                                       self.neural_bounds], axis=0)
                    points = points[in_bound]
                    self.points = np.vstack([self.points, points])
                    self.n_sample += n_sample
                    self.n_reject += n_sample - len(points)
            else:
                try:
                    n_jobs = pool._processes
                except AttributeError:
                    n_jobs = pool.size
                except AttributeError:
                    raise ValueError('Cannot determine size of pool.')
                n_points_per_job = (
                    (max(n_points - len(self.points), 10000)) // n_jobs) + 1
                func = partial(self._reset_and_sample, n_points_per_job)
                rngs = [np.random.default_rng(seed) for seed in
                        np.random.SeedSequence(self.rng.integers(
                            2**32 - 1)).spawn(n_jobs)]
                bounds = pool.map(func, rngs)
                for bound in bounds:
                    self.points = np.vstack([self.points, bound.points])
                    self.n_sample += bound.n_sample
                    self.n_reject += bound.n_reject
                    self.outer_bound.n_sample += bound.outer_bound.n_sample
                    self.outer_bound.n_reject += bound.outer_bound.n_reject

        if return_points:
            points = self.points[:n_points]
            self.points = self.points[n_points:]
            return points

    def volume(self):
        """Return the natural log of the volume.

        Returns
        -------
        log_v : float
            An estimate of the natural log of the volume. Will become more
            accurate as more points are sampled.

        """
        if self.n_sample == 0:
            self.sample(return_points=False)

        return self.outer_bound.volume() + np.log(
            1.0 - self.n_reject / self.n_sample)

    def number_of_networks_and_ellipsoids(self):
        """Return the number of neural networks and sample ellipsoids.

        Returns
        -------
        n_networks : int
            The number of neural networks.
        n_ellipsoids : int
            The number of sample ellipsoids.
        """
        if self.neural_bounds[0].emulator is not None:
            n_networks = len(self.neural_bounds) * len(
                self.neural_bounds[0].emulator.neural_networks)
        else:
            n_networks = 0

        n_ellipsoids = 0
        for bound in self.outer_bound.bounds:
            n_ellipsoids += np.any(~bound.dim_cube)

        return n_networks, n_ellipsoids

    def write(self, group):
        """Write the bound to an HDF5 group.

        Parameters
        ----------
        group : h5py.Group
            HDF5 group to write to.

        """
        group.attrs['type'] = 'NautilusBound'
        group.attrs['n_dim'] = self.n_dim
        group.attrs['n_neural_bounds'] = len(self.neural_bounds)

        for i, neural_bound in enumerate(self.neural_bounds):
            neural_bound.write(group.create_group('neural_bound_{}'.format(i)))

        self.outer_bound.write(group.create_group('outer_bound'))

        group.create_dataset('points', data=self.points,
                             maxshape=(None, self.n_dim))
        group.attrs['n_sample'] = self.n_sample
        group.attrs['n_reject'] = self.n_reject

    def update(self, group):
        """Update bound information previously written to an HDF5 group.

        Parameters
        ----------
        group : h5py.Group
            HDF5 group to write to.

        """
        group.attrs['n_sample'] = self.n_sample
        group.attrs['n_reject'] = self.n_reject
        self.outer_bound.update(group['outer_bound'])
        group['points'].resize(self.points.shape)
        group['points'][...] = self.points

    @classmethod
    def read(cls, group, rng=None):
        """Read the bound from an HDF5 group.

        Parameters
        ----------
        group : h5py.Group
            HDF5 group to write to.
        rng : None or numpy.random.Generator, optional
            Determines random number generation. Default is None.

        Returns
        -------
        bound : NautilusBound
            The bound.

        """
        bound = cls()

        if rng is None:
            bound.rng = np.random.default_rng()
        else:
            bound.rng = rng

        bound.n_dim = group.attrs['n_dim']

        bound.neural_bounds = []
        i = 0
        while 'neural_bound_{}'.format(i) in group:
            bound.neural_bounds.append(NeuralBound.read(
                group['neural_bound_{}'.format(i)],
                rng=bound.rng))
            i += 1

        bound.outer_bound = Union.read(
            group['outer_bound'], rng=rng)

        bound.points = np.array(group['points'])
        bound.n_sample = group.attrs['n_sample']
        bound.n_reject = group.attrs['n_reject']

        return bound

    def reset(self, rng=None):
        """Reset random number generation and any progress, if applicable.

        Parameters
        ----------
        rng : None or numpy.random.Generator, optional
            Determines random number generation. If None, random number
            generation is not reset. Default is None.

        """
        self.points = np.zeros((0, self.n_dim))
        self.n_sample = 0
        self.n_reject = 0
        self.outer_bound.reset(rng)
        if rng is not None:
            self.rng = rng



import multiprocessing
import numpy as np
import os
import pytest
import time

from multiprocessing import Pool

from nautilus import Sampler


def prior(x):
    return x


def likelihood(x):
    time.sleep(0.001)
    return 1, os.getpid()


@pytest.mark.skipif(multiprocessing.get_start_method() == 'spawn',
                    reason=('pytest does not support spawning'))
@pytest.mark.parametrize("pool", [1, 2, Pool(2), Pool(3), None,
                                  (2, 2), (None, 1), (2, Pool(2))])
def test_pool(pool):
    # Test that the expected number of processes are run.

    sampler = Sampler(prior, likelihood, n_dim=2, n_live=50, n_networks=1,
                      pool=pool)
    sampler.run(f_live=1.0, n_eff=0)
    points, log_w, log_l, blobs = sampler.posterior(return_blobs=True)

    if isinstance(pool, tuple):
        pool = pool[0]

    if isinstance(pool, int):
        n_jobs = pool
    elif pool is None:
        n_jobs = 1
    else:
        n_jobs = pool._processes

    assert len(np.unique(blobs)) == n_jobs



"""Modules implementing various multi-dimensional bounds."""

from .basic import UnitCube, Ellipsoid, UnitCubeEllipsoidMixture
from .neural import NeuralBound, NautilusBound
from .union import Union

__all__ = ['UnitCube', 'Ellipsoid', 'UnitCubeEllipsoidMixture', 'Union',
           'NeuralBound', 'NautilusBound']



import numpy as np
import pytest

from multiprocessing import Pool
from scipy.special import gamma

from nautilus import bounds
from nautilus.bounds.basic import minimum_volume_enclosing_ellipsoid
from nautilus.bounds.basic import invert_symmetric_positive_semidefinite_matrix


@pytest.fixture
def points_on_hypersphere_boundary():
    # 2 * n_dim points on the boundary of a unit hypersphere centered on 0.5.

    n_dim = 10
    points = np.zeros((2 * n_dim, n_dim)) + 0.5
    for i in range(n_dim * 2):
        points[i, i // 2] += 1 if i % 2 else -1
    return points


@pytest.fixture
def random_points_from_hypersphere():
    # Random points from a unit hypersphere.

    np.random.seed(0)
    n_dim = 3
    n_points = 1000
    points = np.random.normal(size=(n_points, n_dim))
    points = points / np.sqrt(np.sum(points**2, axis=1))[:, np.newaxis]
    points *= np.random.uniform(size=n_points)[:, np.newaxis]**(1.0 / n_dim)
    return points


@pytest.fixture
def random_points_from_hypercube():
    # Random points from a unit hypersphere.

    np.random.seed(0)
    n_dim = 4
    n_points = 500
    return np.random.random(size=(n_points, n_dim))


def test_unit_cube():
    # Test that UnitCube works as expected.

    n_dim, n_points = 3, 200
    cube = bounds.UnitCube.compute(n_dim)

    points = cube.sample(n_points)
    assert points.shape == (n_points, n_dim)
    assert np.all((points >= 0) & (points < 1))
    assert np.all(cube.contains(points))

    assert cube.volume() == 0


def test_unit_cube_rng():
    # Test that passing a random number generator leads to reproducible
    # results.

    n_dim, n_points = 7, 1000
    cube = bounds.UnitCube.compute(n_dim, rng=np.random.default_rng(0))
    cube_same = bounds.UnitCube.compute(n_dim, rng=np.random.default_rng(0))
    cube_diff = bounds.UnitCube.compute(n_dim, rng=np.random.default_rng(1))
    points = cube.sample(n_points)
    assert np.all(points == cube_same.sample(n_points))
    assert not np.all(points == cube_diff.sample(n_points))
    # We've already sampled from the cube. New samples should be different.
    assert not np.all(points == cube.sample(n_points))

    points = np.random.random((n_points, n_dim))
    assert np.all(cube.contains(points))


def test_invert_symmetric_positive_semidefinite_matrix():
    # Test that the fast implementation of matrix inversion works correctly.

    np.random.seed(0)
    points = np.random.normal(size=(1000, 10))
    m = np.cov(points, rowvar=False)
    assert np.allclose(
        invert_symmetric_positive_semidefinite_matrix(m),
        np.linalg.inv(m))


@pytest.mark.parametrize("tol", [0, 1])
def test_minimum_volume_enclosing_ellipsoid(points_on_hypersphere_boundary,
                                            tol):
    # Test that the MVEE algorithm returns a good approximation to the MVEE.

    c_true = np.median(points_on_hypersphere_boundary, axis=0)
    A_true = np.eye(len(c_true))

    np.random.seed(0)
    # Add a random point. Otherwise, the result is fully correct after the
    # first iteration.
    points = np.concatenate([points_on_hypersphere_boundary,
                             np.atleast_2d(c_true + np.random.random() - 0.5)])
    c, A = minimum_volume_enclosing_ellipsoid(points, tol=tol,
                                              max_iterations=1000)
    assert np.allclose(c, c_true, rtol=0, atol=1e-3) or tol > 0
    assert np.allclose(A, A_true, rtol=0, atol=1e-2) or tol > 0


def test_ellipsoid_construction():
    # Test that the ellipsoid construction fails under certain circumstances.

    with pytest.raises(ValueError):
        bounds.Ellipsoid.compute(np.random.random(size=(10, 10)))

    with pytest.raises(ValueError):
        bounds.Ellipsoid.compute(np.random.random(size=(100, 10)),
                                 enlarge_per_dim=0.9)


def test_ellipsoid_sample_and_contains(points_on_hypersphere_boundary):
    # Test that the ellipsoidal sampling and boundary work as expected.

    ell = bounds.Ellipsoid.compute(
        points_on_hypersphere_boundary, enlarge_per_dim=1.0,
        rng=np.random.default_rng(0))
    c = np.mean(points_on_hypersphere_boundary)

    n_points = 100
    points = ell.sample(n_points)
    assert points.shape == (n_points, points_on_hypersphere_boundary.shape[1])
    assert np.all(np.linalg.norm(points - c, axis=1) < 1 + 1e-9)
    assert np.all(ell.contains(points))

    ell = bounds.Ellipsoid.compute(
        points, enlarge_per_dim=1.1, rng=np.random.default_rng(0))
    points = ell.sample(n_points)
    assert not np.all(np.linalg.norm(points - c, axis=1) < 1)
    assert np.all(ell.contains(points))


def test_ellipsoid_volume(points_on_hypersphere_boundary):
    # Test that the volume of the ellipsoid is accurate.

    for enlarge_per_dim in [1.0, 1.1, np.pi / 2.0]:
        ell = bounds.Ellipsoid.compute(
            points_on_hypersphere_boundary, enlarge_per_dim=enlarge_per_dim)
        n_dim = points_on_hypersphere_boundary.shape[1]
        assert np.isclose(
            ell.volume(), np.log(enlarge_per_dim**n_dim * np.pi**(n_dim / 2) /
                                 gamma(n_dim / 2 + 1)))


def test_ellipsoid_transform(random_points_from_hypersphere):
    # Test that the Cholesky decomposition works correctly.

    ell = bounds.Ellipsoid.compute(
        random_points_from_hypersphere, rng=np.random.default_rng(0))
    points = ell.sample(100)
    points_t = ell.transform(points)
    assert np.all(np.abs(points_t) < 1 + 1e-9)
    assert np.allclose(points, ell.transform(points_t, inverse=True))


def test_ellipsoid_rng(random_points_from_hypersphere):
    # Test that passing a random number generator leads to reproducible
    # results.

    ell = bounds.Ellipsoid.compute(
        random_points_from_hypersphere, rng=np.random.default_rng(0))
    ell_same = bounds.Ellipsoid.compute(
        random_points_from_hypersphere, rng=np.random.default_rng(0))
    ell_diff = bounds.Ellipsoid.compute(
        random_points_from_hypersphere, rng=np.random.default_rng(1))
    n_points = 1000
    points = ell.sample(n_points)
    assert np.all(points == ell_same.sample(n_points))
    assert not np.all(points == ell_diff.sample(n_points))
    assert not np.all(points == ell.sample(n_points))


def test_union_construction():
    # Test that the union construction fails under certain circumstances.

    with pytest.raises(ValueError):
        bounds.Union.compute(np.random.random(size=(100, 10)),
                             n_points_min=5)


def test_union_split(random_points_from_hypersphere):
    # Test that adding ellipsoids works correctly.

    points = np.concatenate([random_points_from_hypersphere,
                             random_points_from_hypersphere + 100,
                             random_points_from_hypersphere + 101])

    union = bounds.Union.compute(
        points, enlarge_per_dim=1.0 + 1e-9, unit=False,
        rng=np.random.default_rng(0))

    # When not allowing overlaps, only 2 ellipsoids should be possible.
    while union.split_bound(allow_overlap=False):
        pass
    assert len(union.bounds) == 2
    assert np.all(union.contains(points))

    # It should be possible to add one more ellipoids when overlaps are
    # allowed.
    assert union.split_bound()
    assert not union.split_bound()
    assert len(union.bounds) == 3
    assert np.all(union.contains(points))


def test_union_split_stops(random_points_from_hypersphere):
    # Test that splitting ellipsoids correctly stops after some time.

    x = np.linspace(-1, 1, 30)
    y = x**2
    points = np.vstack([x, y]).T

    bound = bounds.Union.compute(points)
    n = 0

    while bound.split_bound():
        n += 1

    assert n > 0
    assert np.all(np.array([len(points) for points in bound.points_bounds]) >=
                  bound.n_points_min)


def test_union_sample_and_contains(random_points_from_hypersphere):
    # Test whether the union sampling and boundary work as expected.

    union = bounds.Union.compute(
        random_points_from_hypersphere + 50, enlarge_per_dim=1.0,
        unit=False, rng=np.random.default_rng(0))
    for i in range(4):
        union.split_bound()

    n_points = 100
    points = union.sample(n_points)
    assert points.shape == (n_points, random_points_from_hypersphere.shape[1])
    assert np.all(union.contains(points))

    union_large = bounds.Union.compute(
        points, enlarge_per_dim=1.1, unit=False, rng=np.random.default_rng(0))
    points = union_large.sample(n_points)
    assert not np.all(union.contains(points))


def test_union_rng(random_points_from_hypersphere):
    # Test that passing a random number generator leads to reproducible
    # results.

    union = bounds.Union.compute(
        random_points_from_hypersphere, unit=False,
        rng=np.random.default_rng(0))
    union.split_bound()
    union_same = bounds.Union.compute(
        random_points_from_hypersphere, unit=False,
        rng=np.random.default_rng(0))
    union_same.split_bound()
    union_diff = bounds.Union.compute(
        random_points_from_hypersphere, unit=False,
        rng=np.random.default_rng(1))
    union_diff.split_bound()
    n_points = 100
    points = union.sample(n_points)
    assert np.all(points == union_same.sample(n_points))
    assert not np.all(points == union_diff.sample(n_points))
    assert not np.all(points == union.sample(n_points))


def test_neural_bound_contains(random_points_from_hypercube):
    # Test whether the neural sampling and boundary work as expected.

    points = random_points_from_hypercube
    log_l = -np.linalg.norm(points - 0.5, axis=1)
    log_l_min = np.median(log_l)
    nbound = bounds.NeuralBound.compute(points, log_l, log_l_min, n_networks=1)

    n_points = 100
    n_dim = random_points_from_hypercube.shape[1]
    points = np.random.random(size=(n_points, n_dim))
    log_l = -np.linalg.norm(points - 0.5, axis=1)
    in_bound = nbound.contains(points)
    assert np.mean(log_l[in_bound] > log_l_min) >= 0.9


def test_nautilus_bound_sample_and_contains(random_points_from_hypercube):
    # Test whether the nautilus sampling and boundary work as expected.

    points = random_points_from_hypercube
    log_l = -np.linalg.norm(points - 0.5, axis=1)
    log_l_min = np.median(log_l)
    nbound = bounds.NautilusBound.compute(
        points, log_l, log_l_min, np.log(0.5), n_networks=1)

    n_points = 100
    n_dim = random_points_from_hypercube.shape[1]
    points = np.random.random(size=(n_points, n_dim))
    log_l = -np.linalg.norm(points - 0.5, axis=1)
    in_bound = nbound.contains(points)
    assert np.mean(log_l[in_bound] > log_l_min) >= 0.9

    points = nbound.sample(n_points)
    log_l = -np.linalg.norm(points - 0.5, axis=1)
    assert np.mean(log_l > log_l_min) >= 0.9
    assert np.all(nbound.contains(points))


def test_nautilus_bound_gaussian_shell():
    # Test nautilus sampling on the classic gaussian shell problem. In this
    # problem, the likelihood is high on a thin two-dimensional shell.

    radius = 0.45
    width = 0.01

    np.random.seed(0)
    points = np.random.random((10000, 2))
    log_l = -((np.linalg.norm(points - 0.5, axis=1) - radius) / width)**2
    log_l_min = -1
    points = points[log_l > -100]
    log_l = log_l[log_l > -100]
    log_v_target = np.log(2 * np.pi * radius * width * 2)

    nbound = bounds.NautilusBound.compute(
        points, log_l, log_l_min, log_v_target, split_threshold=1,
        n_networks=1, rng=np.random.default_rng(0))

    points = nbound.sample(10000)
    log_l = -((np.linalg.norm(points - 0.5, axis=1) - radius) / width)**2
    # The volume should be close to the true volume where log_l > log_l_min.
    assert np.isclose(nbound.volume(), log_v_target, rtol=0, atol=np.log(2))
    # Most sampled points should have log_l > log_l_min.
    assert np.mean(log_l > log_l_min) > 0.5
    # We should have only one neural network.
    assert nbound.number_of_networks_and_ellipsoids()[0] == 1


def test_nautilus_bound_two_peaks():
    # Test that the nautilus bound can identify and sample efficienctly from
    # two peaks with wide separations.

    np.random.seed(0)
    radius = 1e-5
    points = np.vstack([np.random.normal(size=(1000, 2)) * radius + 0.1,
                        np.random.normal(size=(1000, 2)) * radius + 0.9])

    def likelihood(x):
        return - np.minimum(
            np.linalg.norm(x - 0.1, axis=-1),
            np.linalg.norm(x - 0.9, axis=-1)) / radius

    log_l = likelihood(points)
    log_l_min = -1
    log_v_target = np.log(2 * np.pi * radius**2)
    nbound = bounds.NautilusBound.compute(
        points, log_l, log_l_min, log_v_target, n_networks=1,
        rng=np.random.default_rng(0))

    points = nbound.sample(10000)
    log_l = likelihood(points)
    # The volume should be close to the true volume where log_l > log_l_min.
    assert np.isclose(nbound.volume(), log_v_target, rtol=0, atol=0.1)
    # Most sampled points should have log_l > log_l_min.
    assert np.mean(log_l > log_l_min) > 0.9
    # We should have two neural networks.
    assert nbound.number_of_networks_and_ellipsoids()[0] == 2


@pytest.mark.parametrize("n_jobs", [1, 2])
def test_nautilus_bound_reset_and_sample(random_points_from_hypercube, n_jobs):
    # Test that resetting the bound works as expected and that we can sample
    # in parallel.

    points = random_points_from_hypercube
    log_l = -np.linalg.norm(points - 0.5, axis=1)
    log_l_min = np.median(log_l)

    if n_jobs > 1:
        pool = Pool(n_jobs)
    else:
        pool = None

    nbound = bounds.NautilusBound.compute(
        points, log_l, log_l_min, np.log(0.5), n_networks=1, pool=pool,
        rng=np.random.default_rng(0))

    nbound.reset(np.random.default_rng(0))
    points_1 = nbound.sample(10000, pool=pool)
    volume_1 = nbound.volume()
    nbound.reset(np.random.default_rng(0))
    points_2 = nbound.sample(10000, pool=pool)
    volume_2 = nbound.volume()

    if n_jobs > 1:
        pool.close()

    assert np.all(points_1 == points_2)
    assert volume_1 == volume_2
